# coding:utf-8
import gc
import time

from catboost import CatBoostClassifier
from lightgbm import LGBMClassifier, LGBMRegressor
import lightgbm as lgb
from scipy.special import inv_boxcox
from scipy.stats import boxcox

from sklearn import metrics
from sklearn.metrics import f1_score, confusion_matrix, accuracy_score, precision_recall_curve, auc, mean_absolute_error
from sklearn.model_selection import StratifiedKFold, KFold
import seaborn as sns
import matplotlib.pyplot as plt
import pandas as pd
import numpy as np
# import statsmodels.api as sm
from sklearn.model_selection import train_test_split
from sklearn.metrics import roc_auc_score
from sklearn.linear_model import LinearRegression
# import pmdarima as pm
# from statsmodels.gam.tests.test_gam import sigmoid
from tqdm import tqdm
import xgboost as xgb

from lianyhaii.metrics import pymetric
from sklearn.model_selection._split import _BaseKFold, indexable, _num_samples
from sklearn.utils.validation import _deprecate_positional_args


def inv_yeojohnson(x, lambda_fitted):
    if lambda_fitted == 0:
        res = np.zeros(len(x))
        for index, value in enumerate(x):
            if value >= 0:
                y = np.exp(value) - 1
            else:
                y = -np.exp(-value) + 1
            res[index] = y
        return res
    if lambda_fitted != 0:
        res = np.zeros(len(x))
        for index, value in enumerate(x):
            if value >= 0:
                y = ((value * lambda_fitted) + 1.0) ** (1.0 / lambda_fitted) - 1.0
            else:
                y = -((-value * (2.0 - lambda_fitted) + 1.0) ** (1.0 / (2.0 - lambda_fitted)) - 1.0)
            res[index] = y
        return res


# https://github.com/getgaurav2/scikit-learn/blob/d4a3af5cc9da3a76f0266932644b884c99724c57/sklearn/model_selection/_split.py#L2243
class GroupTimeSeriesSplit(_BaseKFold):
    """
    Time Series cross-validator variant with non-overlapping groups.
    Provides train/test indices to split time series data samples
    that are observed at fixed time intervals according to a
    third-party provided group.
    In each split, test indices must be higher than before, and thus shuffling
    in cross validator is inappropriate.
    This cross-validation object is a variation of :class:`KFold`.
    In the kth split, it returns first k folds as train set and the
    (k+1)th fold as test set.
    The same group will not appear in two different folds (the number of
    distinct groups has to be at least equal to the number of folds).
    Note that unlike standard cross-validation methods, successive
    training sets are supersets of those that come before them.
    Read more in the :ref:`User Guide <cross_validation>`.
    Parameters
    ----------
    n_splits : int, default=5
        Number of splits. Must be at least 2.
    max_train_size : int, default=None
        Maximum size for a single training set.
    Examples
    --------
    # >>> import numpy as np
    # >>> from sklearn.model_selection import GroupTimeSeriesSplit
    # >>> groups = np.array(['a', 'a', 'a', 'a', 'a', 'a',\
    #                        'b', 'b', 'b', 'b', 'b',\
    #                        'c', 'c', 'c', 'c',\
    #                        'd', 'd', 'd'])
    # >>> gtss = GroupTimeSeriesSplit(n_splits=3)
    # >>> for train_idx, test_idx in gtss.split(groups, groups=groups):
    # ...     print("TRAIN:", train_idx, "TEST:", test_idx)
    # ...     print("TRAIN GROUP:", groups[train_idx],\
    #               "TEST GROUP:", groups[test_idx])
    # TRAIN: [0, 1, 2, 3, 4, 5] TEST: [6, 7, 8, 9, 10]
    # TRAIN GROUP: ['a' 'a' 'a' 'a' 'a' 'a']\
    # TEST GROUP: ['b' 'b' 'b' 'b' 'b']
    # TRAIN: [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10] TEST: [11, 12, 13, 14]
    # TRAIN GROUP: ['a' 'a' 'a' 'a' 'a' 'a' 'b' 'b' 'b' 'b' 'b']\
    # TEST GROUP: ['c' 'c' 'c' 'c']
    # TRAIN: [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14]\
    # TEST: [15, 16, 17]
    # TRAIN GROUP: ['a' 'a' 'a' 'a' 'a' 'a' 'b' 'b' 'b' 'b' 'b' 'c' 'c' 'c' 'c']\
    # TEST GROUP: ['d' 'd' 'd']
    """
    @_deprecate_positional_args
    def __init__(self,
                 n_splits=5,
                 *,
                 max_train_size=None,
                 test_size=None,
                 ):
        super().__init__(n_splits, shuffle=False, random_state=None)
        self.max_train_size = max_train_size
        self.test_size = test_size

    def split(self, X, y=None, groups=None):
        """Generate indices to split data into training and test set.
        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            Training data, where n_samples is the number of samples
            and n_features is the number of features.
        y : array-like of shape (n_samples,)
            Always ignored, exists for compatibility.
        groups : array-like of shape (n_samples,)
            Group labels for the samples used while splitting the dataset into
            train/test set.
        Yields
        ------
        train : ndarray
            The training set indices for that split.
        test : ndarray
            The testing set indices for that split.
        """
        if groups is None:
            raise ValueError(
                "The 'groups' parameter should not be None")
        X, y, groups = indexable(X, y, groups)
        n_samples = _num_samples(X)
        n_splits = self.n_splits
        n_folds = n_splits + 1
        group_dict = {}
        u, ind = np.unique(groups, return_index=True)
        unique_groups = u[np.argsort(ind)]
        n_samples = _num_samples(X)
        n_groups = _num_samples(unique_groups)
        ## 将id分配到group_dic里面。
        for idx in np.arange(n_samples):
            if (groups[idx] in group_dict):
                group_dict[groups[idx]].append(idx)
            else:
                group_dict[groups[idx]] = [idx]

        if n_folds > n_groups:
            raise ValueError(
                ("Cannot have number of folds={0} greater than"
                 " the number of groups={1}").format(n_folds,
                                                     n_groups))
        if self.test_size is None:
            group_test_size = n_groups // n_folds
        else:
            group_test_size = self.test_size

        group_test_starts = range(n_groups - n_splits * group_test_size,
                                  n_groups, group_test_size)

        for group_test_start in group_test_starts:
            train_array = []
            test_array = []
            for train_group_idx in unique_groups[:group_test_start]:
                train_array_tmp = group_dict[train_group_idx]
                train_array = np.sort(np.unique(
                                      np.concatenate((train_array,
                                                      train_array_tmp)),
                                      axis=None), axis=None)

            train_end = train_array.shape
            if self.max_train_size and self.max_train_size < train_end:
                train_array = train_array[train_end -
                                          self.max_train_size:train_end]
            for test_group_idx in unique_groups[group_test_start:
                                                group_test_start +
                                                group_test_size]:
                test_array_tmp = group_dict[test_group_idx]
                concated_test = np.concatenate((test_array,test_array_tmp))
                concated_test = np.unique(concated_test,axis=None)
                test_array = np.sort(concated_test, axis=None)
            yield [int(i) for i in train_array], [int(i) for i in test_array]


class ts_test():
    def __init__(self, tr_df, tt_df, base_features, new_features,
                 m_score, label,
                 ids=None,
                 obj=None,
                 transform_type=None, fitted_param=None,
                 metrices=None,
                 ):

        self.train = tr_df
        self.test = tt_df
        self.base_features = base_features
        self.new_features = new_features
        self.m_score = m_score
        self.label = label
        self.features = base_features + new_features
        self.predictions = None
        self.metrices = metrices
        self.val_pred = None
        self.trn_pred = None
        self.ids = ids
        self.obj = obj
        self.models = []
        self.__feature_importances = None


        self.__transform_type = transform_type
        self.__fitted_param = fitted_param

    def init_CV(self,seed=42, n_split=5,  CV_type='kFold',group_col=None):
        self.cv_conf = {}

        if CV_type == 'lastFold':
            assert not (group_col is None),'lack group col'
            group_block = sorted(self.train[group_col].unique())
            print(group_block)
            ## train_index,test_index
            cv = [[self.train[self.train[group_col] < group_block[-1]].index,
                 self.train[(self.train[group_col] == group_block[-1])].index]]
            self.cv_conf['iter'] = cv
            self.cv_conf['n'] = 1
        elif CV_type == 'kFold':
            cv = KFold(n_splits=n_split, shuffle=True, random_state=seed)
            self.cv_conf['iter'] = cv.split(X=self.train[self.features], y=self.train[self.label])
            self.cv_conf['n'] = n_split
        if CV_type == 'groupTFold':
            assert not (group_col is None),'lack group col'
            cv = GroupTimeSeriesSplit(n_splits=n_split,max_train_size=360,test_size=91)
            self.cv_conf['iter'] = cv.split(X=self.train,groups=self.train[group_col])
            self.cv_conf['n'] = cv.n_splits

        if CV_type == 'submitFold':
            assert not (group_col is None),'lack group col'
            group_block = sorted(self.train[group_col].unique())
            ## train_index,test_index
            ## 使用所有训练集进行提交!
            cv = [[self.train[self.train[group_col] <= group_block[-1]].index,
                 self.train[(self.train[group_col] == group_block[-1])].index]]
            self.cv_conf['iter'] = cv
            self.cv_conf['n'] = 1


    def __check_diff_score(self,y_true,y_pred,):

        pm = pymetric(y_true=y_true,y_pred=y_pred)

        result_score = pm.gen_metric_dict(metric_names=self.metrices,th=0.5)
        for key,value in result_score.items():
            print(f'global {key} : {value}')

        print('='*10+'different with previous version'+'='*10)
        score_list = []
        for n,(key,value) in enumerate(result_score.items()):
            print(f'diff of {key} : {np.round(value-self.m_score[-1][n],5)}')
            score_list.append(value)

        self.m_score.append(score_list)

    def __re_reg(self, x):

        if self.__transform_type == 'ln':
            return np.exp(x) + self.__fitted_param - 1

        elif self.__transform_type == 'boxcox':
            return inv_boxcox(x, self.__fitted_param['lambda']) + self.__fitted_param['min_values'] - 1
        elif self.__transform_type == 'yeojohnson':

            return inv_yeojohnson(x, lambda_fitted=self.__fitted_param)
        elif self.__transform_type == 'my':
            return x * self.__fitted_param
        else:
            return x

    def lgb_test(self, lgb_params, cv_score=False,weight=None):

        ## 用列表存下来用来做验证的index，最后concate在一起。
        cv_score_list = []
        oof_predictions = np.zeros(len(self.train))
        tt_predicts = np.zeros(len(self.test))

        ## 这里预测的是全集
        trn_predicts = np.zeros(len(self.train))

        ## 特征重要性
        imp_df = pd.DataFrame({
            'feat':self.features,
            # 'imp':[]
        })
        # imp_df['score'] = 0
        imp_df['split'] = 0
        imp_df['gain'] = 0

        val_idxs = []
        for n, (trn, val) in enumerate(self.cv_conf['iter']):
            print(f'==== training fold {n+1} ====')
            trn_X, trn_y = self.train.loc[trn, self.features], self.train.loc[trn, self.label]
            val_X, val_y = self.train.loc[val, self.features], self.train.loc[val, self.label]
            val_idxs.append(val)
            if weight is None:
                trn_data = lgb.Dataset(trn_X, label=trn_y,free_raw_data=False)
            else:
                trn_data = lgb.Dataset(trn_X,label=trn_y,weight=self.train.loc[trn,weight])

            val_data = lgb.Dataset(val_X, label=val_y,free_raw_data=False)

            estimator = lgb.train(lgb_params,
                                  trn_data,
                                  valid_sets=[trn_data, val_data],
                                  fobj= self.obj,
                                  verbose_eval=-1,
                                  )
            self.models.append(estimator)

            oof_predictions[val] = self.__re_reg(estimator.predict(val_X))
            imp_df['gain'] = estimator.feature_importance(importance_type='gain')
            imp_df['split'] = estimator.feature_importance(importance_type='split')
            tt_predicts += self.__re_reg(estimator.predict(self.test[self.features])) / self.cv_conf['n']
            trn_predicts += self.__re_reg(estimator.predict(self.train[self.features])) / self.cv_conf['n']

        self.__feature_importances = imp_df.sort_values(['split'],ascending=False,ignore_index=True)
        val = sorted(np.concatenate(val_idxs))
        val_true = self.__re_reg(self.train.loc[val,self.label])
        self.__check_diff_score(y_true=val_true,
                                y_pred=oof_predictions[val])
        self.val_pred = oof_predictions[val]
        self.predictions = tt_predicts
        self.trn_pred = trn_predicts

        if cv_score:
            return oof_predictions, tt_predicts, cv_score_list
        else:
            return oof_predictions, tt_predicts,val_true

    def feature_importances(self,):
        print(self.__feature_importances)
        return self.__feature_importances

    def dl_test(self, model, cv_score=False):

        ## 用列表存下来用来做验证的index，最后concate在一起。
        cv_score_list = []
        oof_predictions = np.zeros(len(self.train))
        tt_predicts = np.zeros(len(self.test))

        ## 这里预测的是全集
        trn_predicts = np.zeros(len(self.train))

        val_idxs = []
        for n, (trn, val) in enumerate(self.cv_conf['iter']):
            print(f'==== training fold {n+1} ====')
            trn_X, trn_y = self.train.loc[trn, self.features], self.train.loc[trn, self.label]
            val_X, val_y = self.train.loc[val, self.features], self.train.loc[val, self.label]
            val_idxs.append(val)

            oof_predictions[val] = self.__re_reg(model.predict(val_X))

            tt_predicts += self.__re_reg(model.predict(self.test[self.features])) / self.cv_conf['n']
            trn_predicts += self.__re_reg(model.predict(self.train[self.features])) / self.cv_conf['n']


        val = np.concatenate(val_idxs)
        val_true = self.__re_reg(self.train.loc[val,self.label])
        self.__check_diff_score(y_true=val_true,
                                y_pred=oof_predictions[val])
        self.val_pred = oof_predictions[val]
        self.predictions = tt_predicts
        self.trn_pred = trn_predicts

        if cv_score:
            return oof_predictions, tt_predicts, cv_score_list
        else:
            return oof_predictions, tt_predicts,val_true

    def xgb_test(self, xgb_params, cv_score=False):
        cv_score_list = []
        oof_predictions = np.zeros(len(self.train))
        tt_predicts = np.zeros(len(self.test))

        for n, (trn, val) in enumerate(self.cv_conf['iter']):
            trn_X, trn_y = self.train.loc[trn, self.features], self.train.loc[trn, self.label]
            val_X, val_y = self.train.loc[val, self.features], self.train.loc[val, self.label]

            tr_data = xgb.DMatrix(trn_X, label=trn_y)
            val_X = xgb.DMatrix(val_X, label=val_y)

            watchlist = [(tr_data, 'train'), (val_X, 'eval')]
            estimator = xgb.train(
                xgb_params,
                tr_data,
                evals=watchlist,
                verbose_eval=1000,
                # num_boost_round=30000,
                # early_stopping_rounds=200,
                early_stopping_rounds=100,
            )

            tt_predicts += estimator.predict(xgb.DMatrix(self.test[self.features])) / self.cv_conf['n']
            oof_predictions[val] = self.__re_reg(estimator.predict(val_X))

            cv_score_list.append(mean_absolute_error(y_true=self.__re_reg(val_y), y_pred=oof_predictions[val]))
            tt_predicts += self.__re_reg(estimator.predict(self.test[self.features])) / self.cv_conf['n']
        self.__check_diff_score(oof_predictions)
        # print(imp.sort_values(['split','gain'],ascending=False,ignore_index=True).loc[imp['feature'].isin(self.new_features),:])
        self.predictions = tt_predicts

        if cv_score:
            return oof_predictions, tt_predicts, cv_score_list
        else:
            return oof_predictions, tt_predicts

    def cat_test(self, cat_params, cv_score=False):
        cv_score_list = []
        oof_predictions = np.zeros(len(self.train))
        tt_predicts = np.zeros(len(self.test))
        cat_features = [x for x in self.train.select_dtypes(include=['category']).columns.tolist() if
                        x in self.features]
        print(cat_features)

        for col in cat_features:
            self.train[col] = pd.to_numeric(self.train[col], errors='ignore', downcast='integer').fillna(-99).astype(
                int)
            self.test[col] = pd.to_numeric(self.test[col], errors='ignore', downcast='integer').fillna(-99).astype(int)

        # print(self.train[cat_features].head())
        from catboost import CatBoostRegressor
        for n, (trn, val) in enumerate(self.cv_conf['iter']):
            trn_X, trn_y = self.train.loc[trn, self.features], self.train.loc[trn, self.label]
            val_X, val_y = self.train.loc[val, self.features], self.train.loc[val, self.label]

            estimator = CatBoostRegressor(**cat_params)
            estimator.fit(
                trn_X, trn_y,
                cat_features=cat_features,
                # early_stopping_rounds=100,
                eval_set=[(val_X,val_y)] if cat_params['task_type'] == 'GPU' else [(trn_X,trn_y),(val_X,val_y)],
                use_best_model=True,
                metric_period=500,
                verbose=True,
            )

            tt_predicts += estimator.predict((self.test[self.features])) / self.cv_conf['n']
            oof_predictions[val] = self.__re_reg(estimator.predict(val_X))

            cv_score_list.append(mean_absolute_error(y_true=self.__re_reg(val_y), y_pred=oof_predictions[val]))
            tt_predicts += self.__re_reg(estimator.predict(self.test[self.features])) / self.cv_conf['n']

        self.__check_diff_score(oof_predictions,)

        self.predictions = tt_predicts
        if cv_score:
            return oof_predictions, tt_predicts, cv_score_list
        else:
            return oof_predictions, tt_predicts

    # @staticmethod
    def plot_model(self,val_true):
        ## 这个不需要分ids
        ## 查看训练集的预测结果与真实值的散点图
        res = pd.DataFrame({
            'pred': self.trn_pred,
            'true': self.__re_reg(self.train[self.label])
        })

        res2 = pd.DataFrame({
            'val_pred': self.val_pred,
            'val_true': val_true,
        })

        figs, axes = plt.subplots(nrows=2, ncols=1, figsize=(10, 10))
        # plt.figure(figsize=(15,15))
        sns.scatterplot(x='true', y='pred', data=res, ax=axes[0])
        axes[0].set_title('scatter of true vs pred in train')
        sns.scatterplot(x='val_true', y='val_pred', data=res2, ax=axes[1])
        axes[1].set_title('scatter of true vs pred in val')
        plt.show()
        del res
        del res2
        gc.collect()
        ## 抽样ids进行趋势查看。
        if self.ids is None:
            ## 查看预测的走势图
            res = pd.DataFrame({
                'pred':np.concatenate((self.trn_pred,self.predictions)),
                'true':np.concatenate((self.__re_reg(self.train[self.label]),self.predictions))
            })

            res.reset_index(drop=False,inplace=True)
            plt.figure(figsize=(20,10))
            sns.lineplot(x='index',y='true',data=res,legend='brief')
            sns.lineplot(x='index', y='pred', data=res, legend='brief', )
            plt.axvline(x=self.train.shape[0],color='black')
            plt.legend()
            plt.show()
        else:
            ids_unique = self.train[self.ids].unique().tolist()
            ## 多个id类型的话需要抽样
            if len(ids_unique) <= 5:
                picked_id = ids_unique
            else:
                picked_id = np.random.choice(ids_unique,5,replace=False)

            res = self.train.copy()
            res['pred'] = self.trn_pred
            res2 = self.test.copy()
            res2['pred'] = self.predictions

            # res2[self.label] = self.__re_reg(res2[self.label])
            res[self.label] = self.__re_reg(res[self.label])

            res = pd.concat([res,res2],ignore_index=True,sort=False)
            del res2
            gc.collect()

            ## 查看预测的走势图
            figs,axes = plt.subplots(nrows=len(picked_id),ncols=1,figsize=(20,10))
            for n,i in enumerate(picked_id):
                tmp = res.loc[res[self.ids]==i]
                tmp_axes = axes[n] if len(picked_id)!=1 else axes
                sns.lineplot(x='date', y='pred', data=tmp,ax=tmp_axes)
                sns.lineplot(x='date', y=self.label, data=tmp,ax=tmp_axes)
                tmp_axes.set_title(f'pred vs true in {i}')
                tmp_axes.legend(['pred','true'])
                tmp_axes.axvline(x=self.train['date'].max(), color='black')
            plt.tight_layout()
            plt.show()

    def submit(self, ID, sub_label):
        today = time.strftime("%Y-%m-%d", time.localtime())[5:]
        self.test[sub_label] = self.predictions
        sub = self.test[[ID, sub_label]].copy()
        plt.figure(figsize=(20, 10))
        sns.distplot(sub[sub_label], bins=100)
        plt.show()
        score = str(np.round(self.m_score[-1][0], 4)) + "_" + str(np.round(self.m_score[-1][1], 4))
        sub.to_csv(f'./result/sub_{today}_{score}.csv', index=False)

    def predict(self,tt_df_features):
        pred = np.zeros(tt_df_features.shape[0])
        for m in self.models:
            pred += self.__re_reg(m.predict(tt_df_features))/len(self.models)
            # pred += m.predict(self.test[self.features])
        return pred


import pandas as pd 
import numpy  as np 
import warnings 
warnings.filterwarnings('ignore')
import sys 
sys.path.append('../')
import lianyhaii

train = pd.DataFrame({
    'y':(np.random.random(size=500)>0.5).astype(int)
})
test = pd.DataFrame({
    'y':(np.random.random(size=500)>0.5).astype(int)
})
for i in range(10):
    train[f'x{i}'] = np.random.random(size=500)
    test[f'x{i}'] = np.random.random(size=500)
label = 'y'

print(train.head())
lgb_params = {
    'objective': 'binary',
    'boosting_type': 'gbdt',
    'metric': 'auc',
    'learning_rate': 0.05,
    'num_leaves': 2**7-1,
    'max_depth': 7,
    'tree_learner': 'serial',
    'colsample_bytree': 0.6,
    'subsample_freq': 2,
    'subsample': 0.6,
    'num_boost_round': 5000,
    # 'max_bin': 100,
    'verbose': -1,
    # 'min_child_samples': 30,
    'seed': 10,
    'early_stopping_rounds': 50,
    'n_jobs': 4,
    'verbose':-1,
}
from lianyhaii.model import *

from sklearn.metrics import precision_score
def top255_precision(true,pred,n=255):
    tmp = pd.DataFrame({
        'pred':pred,
        'true':true,
    })
    print(tmp.mean())
    tmp = tmp.sort_values('pred',ascending=False).head(n)
    tmp['pred'] = 1
    return precision_score(y_true=tmp['true'],y_pred=tmp['pred'])

mt = make_test(train,test,
base_features=[f'x{i}' for i in range(10)],new_features=[],
            m_score=[[0.90913,0.5]],label=label,metrices=['auc',('tp255',top255_precision)],log_tool=None)
mt.init_CV(seed=412,CV_type='skFold',n_split=5)
oof,pred = mt.lgb_test(lgb_params=lgb_params)




import json
import pandas as pd 
import numpy as np
from sklearn.metrics import roc_auc_score
from sklearn.model_selection import train_test_split
from sklearn.preprocessing import LabelEncoder, MinMaxScaler 
import os ,random,gc

from lianyhaii import *
from lianyhaii.model import *
from lianyhaii.feature_tools import *
from lianyhaii.tools import *
from tqdm import tqdm 
tqdm.pandas()


cat_params = {
    'learning_rate': 0.05,
    'bagging_temperature': 0.1,
    'l2_leaf_reg': 30,
    'depth': 12,
    # 'max_leaves': 48,
    'max_bin': 255,
    'iterations': 10000,
    # 'subsample':0.8,
    'task_type': 'GPU',
    'loss_function': "CrossEntropy",
    # 'loss_function': "LogLoss",

    # 'objective': 'MAE',
    'eval_metric': "AUC",
    'bootstrap_type': 'Bayesian',
    # 'bootstrap_type': 'Bernoulli',
    # 'boosting_type': 'Plain',
    'random_seed': 42,
    'early_stopping_rounds': 200,
    'use_best_model': True
}

class encode_din_features(object):
    """
    you should fill na using value "-1"
    hist_feature: [1,2,3,4,5]
    target_feture:1
    """
    def __init__(self,train,test,hist_feature:str,target_feature:str,weight_feature:str) -> None:

        self.train = train 
        self.test = test 
        self.hist_feature = hist_feature
        self.target_feature = target_feature
        self.weight_feature = weight_feature        
    
    @staticmethod
    def _calc_corr(x,y):
        return np.corrcoef(x,y)[0][1]

    def __make(self,size,window,max_len,method):
        sentence = []
        tmp_df = self.train[self.hist_feature].append(self.test[self.hist_feature], )
        target_df = self.train[self.target_feature].append(self.test[self.target_feature], )
        sentence = list(tmp_df)
        print('sample setence')
        print(sentence[0])
        print('start word2vec ...')
        model = Word2Vec(sentence, vector_size=size, window=window, min_count=1,
                         workers=-1, epochs=10, seed=1)

        values = tmp_df.tolist()
        targets = target_df.tolist()
        w2v = []
        for i in tqdm(range(len(values))):
            # if values[i] == '-1':
            #     w2v.append(0)
            #     continue
            cur_hist = [model.wv[x] for x in values[i][:max_len]]
            try :
                cur_target = model.wv[targets[i]]
            except:
                cur_target = [0]*size
            res = list(map(lambda x:self._calc_corr(cur_target,x),cur_hist))
            amean = np.mean(res)
            asum = np.sum(res)
            amax = np.max(res)
            amin = np.min(res)
            astd = np.std(res)
            w2v.append([amean,asum,amax,amin,astd])
        w2v = np.array(w2v)
        np.save('din_encoding.npy',w2v)
        
        ff = []
        for i,f in enumerate(['mean','sum','max','min','std']):
            f_name = f'{self.hist_feature}_{self.target_feature}_din_{f}'
            
            self.train[f_name] = w2v[:len(self.train),i]
            self.test[f_name] = w2v[-len(self.test):,i]
            ff.append(f_name)
        print('finish target encoding')
        return ff
    
    def encode(self,size:int=10,window:int=2,method:str='mean',max_len:int=5):
        return self.__make(size,window,max_len,method) 

def set_seeds(seed=42):
    os.environ['PYTHONHASHSEED'] = str(seed)
    random.seed(seed)
    # tf.random.set_seed(seed)
    np.random.seed(seed)
def clac_gauc(valid):
    auc_score = roc_auc_score(y_true=valid['label'],y_score=valid['pred'])
    print('valid auc',auc_score)
    tmp_cnt = valid.groupby('pvId')['label'].nunique()
    tmp_cnt = tmp_cnt[tmp_cnt==2].index.tolist()
    tmp_valid = valid[valid['pvId'].isin(tmp_cnt)].reset_index(drop=True)
    valid_score = tmp_valid.groupby('pvId').apply(lambda x:roc_auc_score(y_true=x['label'],y_score=x['pred']))
    print('valid gauc',valid_score.mean())


def features_extractor():

    train = pd.read_csv('Sohu2022_data/rec_data/train-dataset.csv')
    test = pd.read_csv('Sohu2022_data/rec_data/test-dataset.csv')
    print(train.shape,test.shape)
    print(train.head())
    test['label'] = -2
    trainb_path= '复赛训练数据/推荐赛道-复赛-训练数据/rec-train-dataset.csv'
    testb_path = '复赛测试数据/rec-test-dataset.csv'
    trainb = pd.read_csv(trainb_path)
    testb = pd.read_csv(testb_path)
    testb['label']  = -1

    test.rename(columns={'testSampleId':'sampleId'},inplace=True)
    testb.rename(columns={'testSampleId':'sampleId'},inplace=True)

    test = pd.concat([test,testb],axis=0,ignore_index=True)
    train = pd.concat([train,trainb],axis=0,ignore_index=True)



    ecf = encode_cat_feats(train,test,base_features+['pvId'],label)
    ecf.encode_lb(inplace=True,method='skl')

    ecf = encode_cat_feats(train,test,['logTs'],label)
    new_time = ecf.encode_lb(inplace=False,method='skl')[0]

    def strings2list(q):
        q = list(map(int,map(lambda x:x.split(':')[0],str(q).split(';'))))
        return q
    from tqdm import tqdm 
    train['userSeq'].fillna('-1',inplace=True)
    test['userSeq'].fillna('-1',inplace=True)
    train['userSeq'] = train['userSeq'].map(strings2list)
    test['userSeq'] = test['userSeq'].map(strings2list)
    ecf = encode_cat_feats(train,test,['userSeq'],label)
    w2v_feats = ecf.encode_w2v()


    data = pd.concat([train,test],axis=0,ignore_index=True)

    data['logTs'] = pd.to_datetime(data['logTs'],unit='ms')
    data['day'] = data['logTs'].dt.day

    ## 当天曝光
    data['day_item_cnt'] = data.groupby(['itemId','day'])['logTs'].transform('count')
    data['day_suv_cnt'] = data.groupby(['suv','day'])['logTs'].transform('count')
    data['day_suv_itemId_cnt'] = data.groupby(['suv','day','itemId'])['logTs'].transform('count')
    data['day_item_nunique'] = data.groupby(['itemId','day'])['logTs'].transform('nunique')
    data['day_suv_nunique'] = data.groupby(['suv','day'])['logTs'].transform('nunique')
    data['day_suv_itemId_nunique'] = data.groupby(['suv','day','itemId'])['logTs'].transform('nunique')


    data['pvId_rank'] = data.groupby('pvId')['logTs'].rank()
    data['suv_rank'] = data.groupby(['pvId', 'suv'])['logTs'].rank()
    data['itemId_rank'] = data.groupby(['pvId', 'itemId'])['logTs'].rank()
    data['rank_nunique'] = data['pvId'].map(data.groupby('pvId')['logTs'].nunique())
    data['suv_nunique'] = data['suv'].map(data.groupby('suv')['logTs'].nunique())
    data['itemId_nunique'] = data['itemId'].map(data.groupby('itemId')['logTs'].nunique())

    cnt_features = []
    for f in tqdm(['pvId','logTs','itemId','suv', 'operator',
                        'browserType',
                        'deviceType',
                        'osType',
                        'province',
                        'city']):
        f_name = f'{f}_cnt'
        data[f_name] = data[f].map(data[f].value_counts())
        cnt_features.append(f_name)

    ## 时间窗口大小
    data['pvId_maxmin'] = data.groupby(['pvId'])['logTs'].transform(lambda x:x.max()-x.min())
    data['pvId_demean'] = data.groupby(['pvId'])['logTs'].apply(lambda x:x-x.mean())
    data['pvId_demean_norm'] = data.groupby(['pvId'])[new_time].apply(lambda x:(x-x.mean())/x.std())

    data.to_feather('Sohu2022_data/rec_data.feather')




def features_extractor2():
    data = pd.read_feather('Sohu2022_data/rec_data.feather')

    data['id_rank'] = data.groupby('pvId')['itemId'].rank()
    data['id_rank_rank'] = data['pvId_rank'] - data['id_rank']

    data['suv_item_cnt'] = data.groupby(['suv','itemId'])['logTs'].transform('count')
    data['suv_item_nunique'] = data.groupby(['suv','itemId'])['logTs'].transform('nunique')
    data['logts_diff'] = data.groupby('pvId')['logTs'].diff().fillna(pd.to_timedelta('0s'))
    data['itemId_suv_rank'] = data.groupby(['suv', 'itemId'])['logTs'].rank()
    data['seq_len'] = data['userSeq'].map(len)
    nuni_feats = ['itemId','pvId',
                'operator',
                'browserType',
                'deviceType',
                'osType',
                'province',
                'city']
    
    for f in nuni_feats:
        data[f'suv_{f}_nuni'] = data.groupby('suv')[f].transform('nunique')
        data[f'suv_{f}_nuni_day'] = data.groupby(['suv','day'])[f].transform('nunique')

    data.to_feather('Sohu2022_data/rec_data.feather')


def features_extractor3():
    data = pd.read_feather('Sohu2022_data/rec_data.feather')

    ## 加入历史序列情感
    res_emo = pd.read_csv('res(1).csv')
    emo_arr = np.array(list(map(list,res_emo['pred'].map(lambda x:np.fromstring(x[1:-1],sep=' ')).values)))
    for i in range(5):
        res_emo[f'{i}_emo'] = emo_arr[:,i]
    emo_f = [f'{i}_emo' for i in range(5)]
    res_emo_stats = res_emo.groupby('id')[emo_f].agg(['mean','std','max','min'])
    emo_feats = [f"agg_{f1}_{f2}" for f1,f2 in res_emo_stats ]
    
    res_emo_stats = res_emo_stats.to_dict()
    emo_hists = []
    for k,v in res_emo_stats.items():
        data[f'emo_hist_{k[0]}_{k[1]}'] = data['userSeq'].progress_map(lambda x:combine_hist(x,v))
        emo_hists.append(f'emo_hist_{k[0]}_{k[1]}')
        # break
    data[emo_hists].to_feather('emo_hists.feather')
    print('emo hist :',emo_hists)


    train,test = data[data['label']!=-1].reset_index(drop=True),data[data['label']==-1].reset_index(drop=True)
    del data  

    edf = encode_din_features(train,test,hist_feature='userSeq',target_feature='itemId',weight_feature='')
    edf.encode(size=5,window=2,max_len=5)
    
    w2v = np.load('din_encoding.npy')
    din_feats= []
    for f in [2,3,4]:
        train[f'din_{f}'] = w2v[:len(train),f]
        test[f'din_{f}'] = w2v[-len(test):,f]
        din_feats.append(f'din_{f}')
    ecf = agg_num_feats(train,test,cat_feats=['pvId'],num_feats=din_feats)
    anf_feats= ecf.agg_feats(agg_type=['max','min','mean','sum'])
    print('din features ',din_feats)
    print('din stat feature ',anf_feats)

    
    res_emo = pd.read_csv('res(1).csv')
    emo_arr = np.array(list(map(list,res_emo['pred'].map(lambda x:np.fromstring(x[1:-1],sep=' ')).values)))
    res_emo['pred'] = res_emo['pred'].map(lambda x:np.fromstring(x[1:-1],sep=' '))
    for i in range(5):
        res_emo[f'{i}_emo'] = emo_arr[:,i]
    emo_f = [f'{i}_emo' for i in range(5)]
    res_emo_stats = res_emo.groupby('id')[emo_f].agg(['mean','std','max','min','sum'])
    emo_feats = [f"agg_{f1}_{f2}" for f1,f2 in res_emo_stats ]
    res_emo_stats.columns = emo_feats
    res_emo_stats.reset_index(drop=False,inplace=True)
    res_emo_stats.rename(columns={'id':'itemId'},inplace=True)
    
    train = train.merge(res_emo_stats,on='itemId',how='left')
    test = test.merge(res_emo_stats,on='itemId',how='left')
    ft = fillna_tools(train,test,emo_feats,label)
    ft.fillna_mean(inplace=True)
    tnf = trans_num_feats(train,test,num_feats=emo_feats,label=label)
    bin_emo_feats = tnf.cut_groups(bins=[20],strategy='kmeans')
    print('bins emo features',bin_emo_feats)

    data = pd.concat([train,test],axis=0,ignore_index=True)
    del train 
    del test 
    
    data.to_feather('Sohu2022_data/rec_data.feather')


def features_extractor4():
    data = pd.read_feather('Sohu2022_data/rec_data.feather')


    data.sort_values('logTs', inplace=True)
    data['logTs_lb_enc'] = data.logTs.map(lambda x:pd.Timestamp(x).timestamp())
    data['log_pv_diff'] = data.groupby(['pvId'])['logTs_lb_enc'].diff()
    df_temp = data.groupby(['pvId'])['log_pv_diff'].agg([
        ('day_range_max', 'max'),
        ('day_range_min', 'min'),
        ('day_range_mean', 'mean'),
        ('day_range_std', 'std'),
        ('day_range_skew', lambda x: x.skew()),
    ])
    diff_features = ['day_range_max','day_range_min','day_range_mean','day_range_std','day_range_skew']
    data = pd.merge(data, df_temp, on='pvId', how='left')

    data['log_suv_diff'] = data.groupby(['suv'])['logTs_lb_enc'].diff()
    df_temp = data.groupby(['suv'])['log_suv_diff'].agg([
        ('suv_range_max', 'max'),
        ('suv_range_min', 'min'),
        ('suv_range_mean', 'mean'),
        ('suv_range_std', 'std'),
        ('suv_range_skew', lambda x: x.skew()),
    ])
    diff_features += ['suv_range_max','suv_range_min','suv_range_mean','suv_range_std','suv_range_skew']
    data = pd.merge(data, df_temp, on='suv', how='left')

    data['log_itemId_diff'] = data.groupby(['itemId'])['logTs_lb_enc'].diff()
    df_temp = data.groupby(['itemId'])['log_itemId_diff'].agg([
        ('itemId_range_max', 'max'),
        ('itemId_range_min', 'min'),
        ('itemId_range_mean', 'mean'),
        ('itemId_range_std', 'std'),
        ('itemId_range_skew', lambda x: x.skew()),
    ])
    diff_features += ['itemId_range_max','itemId_range_min','itemId_range_mean','itemId_range_std','itemId_range_skew']
    data = pd.merge(data, df_temp, on='itemId', how='left')
    

    
    data.to_feather('Sohu2022_data/rec_data.feather')


def features_extractor5():
    
    data = []
    f = open('Sohu2022_data/rec_data/recommend_content_entity_0317.txt',encoding='utf-8',mode='r')
    for line in tqdm(f):
        line = json.loads(line)
        id = line['id']
        entity = ",".join(line['entity'])
        data.append([id,entity])


    f = open('复赛测试数据/recommend_content_entity.txt',encoding='utf-8',mode='r')
    for line in tqdm(f):
        line = json.loads(line)
        id = line['id']
        entity = ",".join(line['entity'])
        data.append([id,entity])


    f = open('复赛训练数据/recommend_content_entity.txt',encoding='utf-8',mode='r')
    for line in tqdm(f):
        line = json.loads(line)
        id = line['id']
        entity = ",".join(line['entity'])
        data.append([id,entity])
    data = pd.DataFrame(data,columns=['id','entity'])

    from sklearn.feature_extraction.text import TfidfVectorizer

    tfv = TfidfVectorizer(max_features=5000,)
    res = tfv.fit_transform(data['entity'])
    from sklearn.decomposition import TruncatedSVD

    tsvd = TruncatedSVD(n_components=30,random_state=1003)
    res_svd = tsvd.fit_transform(res)
    for f in range(30):
        data[f'tfidf_svd_{f}'] = res_svd[:,f]

    data['id'] = data['id'].astype(int)
    data.to_feather('entity_svd_data.feather')

def combine_hist(x,dict):
    xs = []
    for f in x :
        if f in dict:
            xs.append(dict[f])
        else:
            xs.append(0)
    return sum(xs)

if __name__ == '__main__':
    label = 'label'

    base_features = [
                'suv',
                # 'itemId',
                'operator',
                'browserType',
                'deviceType',
                'osType',
                'province',
                'city']
    features_extractor()
    features_extractor2()
    features_extractor3()
    features_extractor4()
    features_extractor5()
    data = pd.read_feather('Sohu2022_data/rec_data.feather')
    remove_features= ['sampleId','label','logTs','pvId','userSeq']


    new_features = [
    'day_item_cnt','day_suv_cnt','pvId_rank','suv_rank','itemId_rank','rank_nunique',
    'suv_nunique','itemId_nunique','day_suv_itemId_cnt',
        'pvId_cnt', 'logTs_cnt', 'itemId_cnt', 'suv_cnt', 'operator_cnt', 'browserType_cnt', 'deviceType_cnt', 'osType_cnt', 'province_cnt', 'city_cnt',
        'userSeq_W2V_0', 'userSeq_W2V_1', 'userSeq_W2V_2', 'userSeq_W2V_3', 'userSeq_W2V_4', 'userSeq_W2V_5', 'userSeq_W2V_6', 'userSeq_W2V_7', 'userSeq_W2V_8', 'userSeq_W2V_9',

    ]+['pvId_demean','pvId_maxmin','pvId_demean_norm']\
        +['pvId','itemId']+['id_rank','id_rank_rank',]\

    new_features += ['suv_item_cnt','suv_item_nunique','logts_diff']

    ## 对用户的刻画
    new_features += ['itemId_suv_rank','seq_len']
    nuni_feats = ['itemId','pvId',
                'operator',
                'browserType',
                'deviceType',
                'osType',
                'province',
                'city']

    cnt_features = []
    for f in nuni_feats:
        cnt_features.extend([f'suv_{f}_nuni',f'suv_{f}_nuni_day'])
    new_features += cnt_features

    diff_features = ['day_range_max', 'day_range_min', 'day_range_mean', 'day_range_std', 'day_range_skew', 'suv_range_max', 'suv_range_min', 'suv_range_mean', 'suv_range_std', 'suv_range_skew', 'itemId_range_max', 'itemId_range_min', 'itemId_range_mean', 'itemId_range_std', 'itemId_range_skew',]
    new_features += diff_features


    stats_din_featurs = ['pvId_din_2_agg_max', 'pvId_din_2_agg_min', 'pvId_din_2_agg_mean', 'pvId_din_2_agg_sum', 'pvId_din_3_agg_max', 'pvId_din_3_agg_min', 'pvId_din_3_agg_mean', 'pvId_din_3_agg_sum', 'pvId_din_4_agg_max', 'pvId_din_4_agg_min', 'pvId_din_4_agg_mean', 'pvId_din_4_agg_sum']
    din_feats = ['din_2', 'din_3', 'din_4']
    new_features += stats_din_featurs+din_feats

    emo_hists = ['emo_hist_0_emo_mean', 'emo_hist_0_emo_std', 'emo_hist_0_emo_max', 'emo_hist_0_emo_min', 'emo_hist_1_emo_mean', 'emo_hist_1_emo_std', 'emo_hist_1_emo_max', 'emo_hist_1_emo_min', 'emo_hist_2_emo_mean', 'emo_hist_2_emo_std', 'emo_hist_2_emo_max', 'emo_hist_2_emo_min', 'emo_hist_3_emo_mean', 'emo_hist_3_emo_std', 'emo_hist_3_emo_max', 'emo_hist_3_emo_min', 'emo_hist_4_emo_mean', 'emo_hist_4_emo_std', 'emo_hist_4_emo_max', 'emo_hist_4_emo_min']
    bin_emo_feats = ['agg_0_emo_mean_20_bin_kmeans', 'agg_0_emo_std_20_bin_kmeans', 'agg_0_emo_max_20_bin_kmeans', 'agg_0_emo_min_20_bin_kmeans', 'agg_0_emo_sum_20_bin_kmeans', 'agg_1_emo_mean_20_bin_kmeans', 'agg_1_emo_std_20_bin_kmeans', 'agg_1_emo_max_20_bin_kmeans', 'agg_1_emo_min_20_bin_kmeans', 'agg_1_emo_sum_20_bin_kmeans', 'agg_2_emo_mean_20_bin_kmeans', 'agg_2_emo_std_20_bin_kmeans', 'agg_2_emo_max_20_bin_kmeans', 'agg_2_emo_min_20_bin_kmeans', 'agg_2_emo_sum_20_bin_kmeans', 'agg_3_emo_mean_20_bin_kmeans', 'agg_3_emo_std_20_bin_kmeans', 'agg_3_emo_max_20_bin_kmeans', 'agg_3_emo_min_20_bin_kmeans', 'agg_3_emo_sum_20_bin_kmeans', 'agg_4_emo_mean_20_bin_kmeans', 'agg_4_emo_std_20_bin_kmeans', 'agg_4_emo_max_20_bin_kmeans', 'agg_4_emo_min_20_bin_kmeans', 'agg_4_emo_sum_20_bin_kmeans']
    new_features += emo_hists+bin_emo_feats
    emo_hists_df = pd.read_feather('emo_hists.feather')
    data = pd.concat([data,emo_hists_df],axis=1)
    del emo_hists_df 
    gc.collect()

    w2v_feats = ['userSeq_W2V_0', 'userSeq_W2V_1', 'userSeq_W2V_2', 'userSeq_W2V_3', 'userSeq_W2V_4', 'userSeq_W2V_5', 'userSeq_W2V_6', 'userSeq_W2V_7', 'userSeq_W2V_8', 'userSeq_W2V_9',]
    dense_features = diff_features + ['itemId_suv_rank','seq_len',] + cnt_features+emo_hists+stats_din_featurs+din_feats
    data[dense_features] = data[dense_features].fillna(-1)

    entity_df = pd.read_feather('entity_svd_data.feather')
    entity_df.rename(columns={'id':'itemId'},inplace=True)
    del entity_df['entity']
    gc.collect()

    entity_df = entity_df.groupby('itemId').max().reset_index(drop=False)
    entity_feats = entity_df.drop(['itemId',],axis=1).columns.tolist()
    
    data = data.merge(entity_df,on='itemId',how='left')
    data[entity_feats] = data[entity_feats].fillna(-1)
    new_features += entity_feats
    del entity_df
    gc.collect()

    data = data[data['label']!=-2].reset_index(drop=True)
    train,test = data[data['label']!=-1].reset_index(drop=True),data[data['label']==-1].reset_index(drop=True)
    del data 
    gc.collect()
    
    time_gap = '20220104160000'
    trn_mask = train['logTs']<=time_gap
    val_mask = train['logTs']>time_gap 

    train.loc[trn_mask,'fold'] = 0 
    train.loc[~trn_mask,'fold'] = 1 


    cat_params['learning_rate'] = 0.01
    cat_params['iterations'] = int(2600)
    cat_params['early_stopping_rounds'] = 10000
    mt = make_test(train,test,
    base_features=base_features,new_features=new_features,
            m_score=[[0.5,0.5]],label=label,metrices=['auc','f1'],log_tool=None)
    mt.init_CV(seed=42,CV_type='gFold',n_split=5,group_col='pvId')
    oof,pred = mt.cat_test(cat_params=cat_params,save_path='.')
    # train.loc[val_mask,'pred'] = oof 
    # clac_gauc(train[val_mask].reset_index(drop=True))

    version = 'submission'
    from pathlib import Path 
    Path(version).mkdir(exist_ok=True,parents=True)
    sub = test[['sampleId']]
    sub['Id'] = sub['sampleId'].astype(int)
    sub['result'] = pred.reshape(-1)
    sub[['Id','result']].to_csv(f'{version}/section2.txt',sep='\t',index=False)


# coding:utf-8

import random
import os
import numpy as np
import pandas as pd

from sklearn.metrics import roc_auc_score, f1_score, mean_squared_error, mean_absolute_error, \
    mean_absolute_percentage_error

# def metric_zs(y_true,y_pred):

def mape_zx(y_true,y_pred):
    assert len(y_true) == len(y_pred),'lenght must be the same'
    ## 排除ytrue中的缺失情况
    ytruena = np.isnan(y_true)
    y_pred = y_pred[~ytruena]
    y_true = y_true[~ytruena]
    ## demo
    # y_pred = np.array([1, np.nan, 3,4,5,6,9])
    # y_true = np.array([2, 3, 5,4.8,5.8,6.1,np.nan])
    w = np.abs(y_true - y_pred) / (y_true)
    na_ids = np.isnan(y_pred)
    max_ids = (w>=0.2)&(~na_ids)
    min_ids = (w<0.2)&(~na_ids)
    w[min_ids] = 100
    w[max_ids] = 0
    w[na_ids] = 40

    return w.sum()/w.shape[0]

def mape_zx_lgb(y_true,y_pred):
    assert len(y_true) == len(y_pred),'lenght must be the same'
    w = np.abs(y_true - y_pred) / (y_true)
    max_ids = w>=0.2
    min_ids = w<0.2
    w[min_ids] = 100
    w[max_ids] = 0

    return w.sum()/w.shape[0]

def mape_obj(y_pred,data):
    y_true = data.get_label()
    w = y_true
    # w[y_true<0.001] = w[y_true<0.001]**0.0001
    grad = np.sign(y_pred - y_true)/(w)

    hess = 1/(w)
    # hess = np.zeros_like(y_true)
    # grad[y_true < 0.01] *= 0.5
    # hess[y_true < 0.01] *= 0.5

    grad[(y_true==0)] = 0
    hess[(y_true==0)] = 0

    return grad,hess

def mape_zs_task1(y_pred,data):
    data['pred'] = y_pred
    ## 删除岗位B的非工作日记录
    tmp = data[~((data['post_id']=='B')&(~data['WKD_TYP_CD'].isin(['WN','WS'])))]
    # tmp = data[~((data['post_id']=='B')&(~data['WKD_TYP_CD']==0))]

    score = np.abs(tmp['pred'] - tmp['amount'])/(tmp['amount']+1)
    return score.sum()/score.shape[0]

def mape_zs_task2(y_pred,data):
    data['pred'] = y_pred
    ## 删除岗位B的非工作日记录
    tmp = data[~((data['post_id']=='B')&(~data['WKD_TYP_CD'].isin(['WN','WS'])))]
    ## 删除非工作时间的数据
    tmp['hour'] = tmp['date'].dt.strftime('%H%M')
    tmp = tmp[tmp['hour'] > '0830']
    tmp = tmp[tmp['hour'] < '1830']
    tmp.reset_index(drop=True,inplace=True)

    score = np.abs(tmp['pred'] - tmp['amount'])/(tmp['amount']+1)
    return score.sum()/score.shape[0]


def tpr_weight_funtion(y_true,y_predict):
    d = pd.DataFrame()

    d['prob'] = list(y_predict)

    d['y'] = list(y_true)

    d = d.sort_values(['prob'], ascending=[0])

    y = d.y

    PosAll = pd.Series(y).value_counts()[1]

    NegAll = pd.Series(y).value_counts()[0]

    pCumsum = d['y'].cumsum()

    nCumsum = np.arange(len(y)) - pCumsum + 1

    pCumsumPer = pCumsum / PosAll

    nCumsumPer = nCumsum / NegAll

    TR1 = pCumsumPer[abs(nCumsumPer-0.001).idxmin()]

    TR2 = pCumsumPer[abs(nCumsumPer-0.005).idxmin()]

    TR3 = pCumsumPer[abs(nCumsumPer-0.01).idxmin()]
    # print(f'0.1% tpr is {TR1}, 0.5% tpr is {TR2} ,1% tpr is {TR3}')
    return 0.4 * TR1 + 0.3 * TR2 + 0.3 * TR3
# def mape(y_true, y_pred):
#     return np.mean(np.abs((y_pred - y_true) / y_true)) * 100


def smape(y_true, y_pred):
    return 2.0 * np.mean(np.abs(y_pred - y_true) / (np.abs(y_pred) + np.abs(y_true))) * 100


class pymetric():
    def __init__(self,y_true,y_pred,):
        # self.metric_name = metric_name
        self.yt = y_true
        self.yp = y_pred
        self.tt = {}
        self.new_metrics = []
    def base_metric(self,metric_est,threshold=None):
        """

        :param metric_est: a estimtor to score
        # :param direct:
        :param threshold:
        :return: int of score!
        """
        if not threshold:
            return metric_est(self.yt,self.yp)
        else:
            ytmp = [int(x>threshold) for x in self.yp]
            return metric_est(self.yt,ytmp)
    def add_metric(self,metric_name,metric_fuction):

        self.new_metrics.append((metric_name,metric_fuction))

    @staticmethod
    def recall_score(y_true,y_pred):
        yt = np.array(y_true)
        yp = np.array(y_pred)
        p_sum = yt.sum()
        tp = yp[yt == 1].sum()
        return tp / (p_sum + 1e-10)

    @staticmethod
    def acc_score(y_true,y_pred):
        yt = np.array(y_true)
        yp = np.array(y_pred)
        p_sum = yp.sum()
        tp = yp[yt == 1].sum()
        return tp / (p_sum + 1e-10)

    def gen_metric_dict(self,metric_names,th=0.5):
        for name in metric_names:
            if name == 'auc':
                self.tt['auc'] = self.base_metric(roc_auc_score)
            if name == 'f1':
                self.tt['f1'] = self.base_metric(f1_score,threshold=th)
            if name == 'recall':
                self.tt['recall'] = self.base_metric(self.recall_score,threshold=th)
            if name == 'acc':
                self.tt['acc'] = self.base_metric(self.acc_score,threshold=th)
            if name == 'tpr':
                self.tt['tpr'] = self.base_metric(tpr_weight_funtion)

            if name == 'mse':
                self.tt['mse'] = self.base_metric(mean_squared_error)
            if name == 'rmse':
                self.tt['rmse'] = np.sqrt(self.base_metric(mean_squared_error))
            if name == 'mae':
                self.tt['mae'] = self.base_metric(mean_absolute_error)
            if name == 'mape':
                self.tt[name] = self.base_metric(mean_absolute_percentage_error)
            if name == 'smape':
                self.tt[name] = self.base_metric(smape)
            if name == 'mape_zx':
                self.tt[name] = self.base_metric(mape_zx)
            for n,f in self.new_metrics:
                if name == n:
                    self.tt[n] = self.base_metric(f)
            # if name == 'mape_zs_task1':
            #     self.tt[name] = self.base_metric(mape_zx)

        return self.tt







from itertools import combinations
from matplotlib.pyplot import plot
import pandas as pd 
import numpy as np 

import lianyhaii
from sklearn import feature_selection

from lianyhaii.model import *
from lianyhaii.fe_selector import *
from lianyhaii.eda_tool import *
from lianyhaii.tools import * 


lgb_params = {
    'objective': 'binary',
    'boosting_type': 'gbdt',
    'metric': 'auc',
    'learning_rate':0.01,
    'colsample_bytree':0.7,
    'subsample':0.7,
    'num_leaves': 2**7-1,
    'max_depth': 7,
    'tree_learner': 'serial',
    'max_bin': 155,
    'min_child_samples': 30,
    'early_stopping_rounds':100,
    'num_boost_round':5000,
    'reg_lambda':0.5,
    'reg_alpha':0.5,
    'seed':42,
    'n_jobs': -1,
    'verbose':-1,
}

train = pd.read_csv('data/train.csv')
test = pd.read_csv('data/test_B榜.csv')
label = "LABEL"
test[label] = -1
remove_features = ['CUST_UID',label]
total_features = [x for x in train if x not in remove_features]

data = pd.concat([train,test],axis=0,ignore_index=True)

abcd_feats = ['NB_CTC_HLD_IDV_AIO_CARD_SITU',
'WTHR_OPN_ONL_ICO','LGP_HLD_CARD_LVL']
aio_map ={
    'A':1,    'B':2,    'C':3,    'D':4,    'E':5,    'F':6,    '?':-999,
}
yes_map = {    'Y':1,    '?':-999,}
yes_feats = ['MON_12_CUST_CNT_PTY_ID']
for f in abcd_feats:
    data[f] =data[f].map(aio_map)
data['MON_12_CUST_CNT_PTY_ID'] = data['MON_12_CUST_CNT_PTY_ID'].map(yes_map)
remove_features = ['CUST_UID',label]
for f in data.columns:
    if f in remove_features:continue
    data[f] = pd.to_numeric(data[f],errors='coerce',) - 2
train = data[data[label]!=-1].reset_index(drop=True)
test = data[data[label]==-1].reset_index(drop=True)
from lianyhaii.feature_tools import *
tnf = trans_num_feats(train,test,num_feats=[x for x in total_features if x not in abcd_feats+yes_feats],label=label)
tnf.log_num_feats(inplace=True)
total_features =  [x for x in train if x not in remove_features]
fs = FeatureSelected(train,test,feats=total_features,label=label)
# # fs.std_selector(plot_it=True)
# fs.adv_test_selector(plot_it=False,threshold=0.05)
# # fs.ks_test_selector(plot_it=True)
fs.remove_by_all()
fs.tt_drop_feats += ['AGN_CUR_YEAR_WAG_AMT', 'CUR_YEAR_PUB_TO_PRV_TRX_PTY_CNT','CUR_MON_EXT_SAM_CUST_TRSF_OUT_AMT','CUR_YEAR_COUNTER_ENCASH_CNT', 'OPN_TM', 'REG_DT', 'ICO_CUR_MON_ACM_TRX_TM', 'CUR_MON_COR_DPS_MON_DAY_AVG_BAL','CUR_MON_EXT_SAM_CUST_TRSF_IN_AMT','NB_RCT_3_MON_LGN_TMS_AGV','AGN_AGR_LATEST_AGN_AMT',"REG_CPT",'COUNTER_CUR_YEAR_CNT_AMT', 'COR_KEY_PROD_HLD_NBR', 'CUR_YEAR_MON_AGV_TRX_CNT', 'CUR_YEAR_MID_BUS_INC','CUR_YEAR_COR_DPS_YEAR_DAY_AVG_INCR','AI_STAR_SCO','EMP_NBR', 'PUB_TO_PRV_TRX_AMT_CUR_YEAR','CUR_YEAR_COR_DMND_DPS_DAY_AVG_BAL', 'AGN_CUR_YEAR_AMT', 'ICO_CUR_MON_ACM_TRX_AMT', 'WTHR_OPN_ONL_ICO',]
lgb_params['boosting_type'] = 'rf'
lgb_params['bagging_freq'] = 1
lgb_params['bagging_fraction'] = 0.8
mt = make_test(train,test,base_features=[x for x in total_features if x not in fs.tt_drop_feats],new_features=[],
            m_score=[[0.,0.]],label=label,metrices=['auc','f1'],log_tool=None)
mt.init_CV(seed=42,CV_type='skFold',n_split=5)
oof,pred = mt.lgb_test(lgb_params=lgb_params)
sub = test[['CUST_UID']].copy()
sub['pred'] = mt.predictions
sub.to_csv('submission.txt',index=False,header=False,sep='\t')


# coding:utf-8


import functools
import random
import os
import sys
import time

import numpy as np
import pandas as pd
from scipy.stats import yeojohnson,boxcox
from sklearn.metrics import f1_score


# from autograd import elementwise_grad as egrad
# import autograd.numpy as npe

from lianyhaii.feature_tools import fillna_tools

import logging
from logging import handlers


# def f1_macro_(y_pred, y_true, n_labels=2):
#     total_f1 = 0.
#     for i in range(n_labels):
#         yt = y_true == i
#         yp = y_pred == i

#         tp = np.sum(yt & yp)

#         tpfp = npe.sum(yp)
#         tpfn = npe.sum(yt)
#         if tpfp == 0:
#             # print('[WARNING] F-score is ill-defined and being set to 0.0 in labels with no predicted samples.')
#             precision = 0.
#         else:
#             precision = tp / tpfp
#         if tpfn == 0:
#             # print(f'[ERROR] label not found in y_true...')
#             recall = 0.
#         else:
#             recall = tp / tpfn

#         if precision == 0. or recall == 0.:
#             f1 = 0.
#         else:
#             f1 = 2 * precision * recall / (precision + recall)
#         total_f1 += f1
#     return total_f1 / n_labels


# def f1_macro_loss(preds,data):
#     # print(type(preds),type(data))
#     y_label = npe.array(data.get_label())
#     preds = npe.array([int(x>0.5) for x in preds])

#     grad_f1 = egrad(f1_macro_,1)
#     hessian_f1 = egrad(egrad(f1_macro_,1))
#     grad_f1_score = grad_f1(preds,y_label)
#     hessian_f1_score = hessian_f1(preds,y_label)

#     return grad_f1_score,hessian_f1_score

def seed_everything(seed=0):
    random.seed(seed)
    os.environ['PYTHONHASHSEED'] = str(seed)
    np.random.seed(seed)

def load_reg_params(seed):
    lgb_params = {
        'objective': 'rmse',
        'boosting_type': 'gbdt',
        # 'boosting_type': 'goss',
        'metric': 'rmse',
        'n_jobs': -1,
        'learning_rate': 0.05,
        'num_leaves': 2**5-1,
        # 'max_depth': 5,
        'tree_learner': 'serial',
        'colsample_bytree': 0.8,
        'subsample_freq': 1,
        'subsample': 0.8,
        'num_boost_round': 1000,
        # 'max_bin': 100,
        'verbose': -1,
        # 'min_child_samples': 30,
        'seed': seed,
        # 'early_stopping_rounds': 100,
    }

    xgb_params = {
        'objective': 'reg:squarederror',
        # 'objective':'binary:logitraw',
        # 'objective':'binary:hinge',
        #     'booster': 'GPU_hist',
        'eval_metric': 'mae',
        #     'eval_metric': 'auc',

        'n_estimators':10000,
        'tree_method': 'hist',
        'max_depth': 8,
        'subsample': 0.8,
        'colsample_bytree': 0.8,
        'eta': 0.01,
        'seed': seed,
        'nthread': -1,
        # 'silent': True,
    }

    cat_params = {
        'learning_rate': 0.1,
        'bagging_temperature': 0.1,
        'l2_leaf_reg': 30,
        'depth': 12,
        # 'max_leaves': 48,
        'max_bin': 255,
        'iterations': 10000,
        'task_type': 'GPU',
        'loss_function': "MAE",
        # 'objective': 'MAE',
        'eval_metric': "MAE",
        'bootstrap_type': 'Bayesian',
        'random_seed': seed,
        'early_stopping_rounds': 100,
        'use_best_model': True
    }

    lr_params = {
        'random_state': seed,
        'C': 1,
        'max_iter': 1000,
        'n_jobs': -1,

    }

    model_params = {
        'lgb': lgb_params,
        'xgb': xgb_params,
        'cat':cat_params,
        'lr':lr_params,
    }

    return model_params

def time_it(fun):
    @functools.wraps(fun)
    def wrapper(*args,**kwargs):
        """this is wrapper function"""
        func_name = fun.__name__
        start_time = time.time()
        res = fun(*args,**kwargs)
        end_time = time.time()
        print(f'{func_name} costing time : {(end_time-start_time):.2f}' )
        # print(f'costing time :{(end_time-start_time)}' )
        return res
    return wrapper

def add_base_features(test_feats=True, save_feats=False):
    global train, test
    global feature_path,ID
    new_feats = []
    file_name = sys._getframe().f_code.co_name[4:]
    if (not test_feats) and (not save_feats):
        ## 读取数据
        tr_feats = pd.read_feather(feature_path + f'train_feats_{file_name}.feather')
        tt_feats = pd.read_feather(feature_path + f'test_feats_{file_name}.feather')
        feats = [x for x in tr_feats if x not in [ID]]
        ## 合并数据
        train = train.merge(tr_feats, on=ID, how='left')
        test = test.merge(tt_feats, on=ID, how='left')
        return feats

    tt_new_features = [x for item in new_feats for x in item]
    if (not test_feats) and save_feats:
        train[[ID] + tt_new_features].to_feather(feature_path + f'train_feats_{file_name}.feather')
        test[[ID] + tt_new_features].to_feather(feature_path + f'test_feats_{file_name}.feather')
        print('saved feats', file_name)
    if test_feats:
        new_feats.append(tt_new_features)
        return new_feats
    else:
        return tt_new_features

def reduce_mem_usage(df, verbose=True):
    numerics = ['int16', 'int32', 'int64', 'float16', 'float32', 'float64']
    start_mem = df.memory_usage().sum() / 1024 ** 2
    for col in df.columns:
        col_type = df[col].dtypes
        if col_type in numerics:
            c_min = df[col].min()
            c_max = df[col].max()
            if str(col_type)[:3] == 'int':
                if c_min > np.iinfo(np.int8).min and c_max < np.iinfo(np.int8).max:
                    df[col] = df[col].astype(np.int8)
                elif c_min > np.iinfo(np.int16).min and c_max < np.iinfo(np.int16).max:
                    df[col] = df[col].astype(np.int16)
                elif c_min > np.iinfo(np.int32).min and c_max < np.iinfo(np.int32).max:
                    df[col] = df[col].astype(np.int32)
                elif c_min > np.iinfo(np.int64).min and c_max < np.iinfo(np.int64).max:
                    df[col] = df[col].astype(np.int64)
            else:
                if c_min > np.finfo(np.float16).min and c_max < np.finfo(np.float16).max:
                    df[col] = df[col].astype(np.float16)
                elif c_min > np.finfo(np.float32).min and c_max < np.finfo(np.float32).max:
                    df[col] = df[col].astype(np.float32)
                else:
                    df[col] = df[col].astype(np.float64)
    end_mem = df.memory_usage().sum() / 1024 ** 2
    if verbose:
        print('Mem. usage decreased to {:5.2f} Mb ({:.1f}% reduction)'.format(
            end_mem, 100 * (start_mem - end_mem) / start_mem))
    return df

def transform_label(train,label,transform_type='ln',params=None):

    fitted_lambda = None
    if transform_type == 'ln':
        f_label = f"ln_{label}"
        train[f_label] = np.log(train[label]-train[label].min()+1)
        fitted_lambda = train[label].min()
        # return fitted_lambda
    elif transform_type == 'boxcox':
        f_label = f"bc_{label}"
        train[f_label] ,fitted_lambda= boxcox(train[label]-train[label].min()+1)
        fitted_lambda ={
            'lambda':fitted_lambda,
            'min_values':train[label].min(),
        }

    elif transform_type == 'yeojohnson':
        f_label = f"yj_{label}"
        train[f_label],fitted_lambda = yeojohnson(train[label],)
    elif transform_type == 'my':
        smoothing = 1e4
        f_label = f'my_{label}'
        train[f_label] = train[label] / smoothing
        fitted_lambda = smoothing

    elif transform_type == 'minmax':
        smoothing = 15
        f_label = f'my_{label}'
        train[f_label] = train[label].clip(0,smoothing) / smoothing
        fitted_lambda = smoothing
    else:
        f_label = label
        pass
    label = f_label
    return label,fitted_lambda


def clean_null(test_feats=True,label=None):
    global train,test
    new_feats = []
    tt_cols =  ['X38']
    ft = fillna_tools(train,test,tt_cols,label)
    ff = ft.fillna_min()
    new_feats.append(ff)
    ff = ft.fillna_mean()
    new_feats.append(ff)
    ff = ft.fillna_median()
    new_feats.append(ff)
    ff = ft.fillna_mode()
    new_feats.append(ff)
    new_feats.append(tt_cols)

    tt_new_features = [x for item in new_feats for x in item]
    if test_feats:
        new_feats.append(tt_new_features)
        return new_feats
    else:
        return tt_new_features

class resample_tools:

    def __init__(self,train:pd.DataFrame,test:pd.DataFrame,group:str,label:str):
        self.train = train
        self.test = test
        self.group = group

    def StratifiedSample(self):
        """ 使得他们的分层比例一样"""
        tr_group = self.train[self.group].unique()
        tt_group = self.test[self.group].unique()

        tr_size = self.train.shape[0]
        tt_size = self.test.shape[0]

        keep_tr = []

        for n in tt_group:
            ## 如果n不在train中则跳过
            if n not in tr_group:
                continue
            ## 如果tt中比例高于tr也跳过,但是保留所有索引
            tt_ratio = self.test[self.test[self.group] == n].shape[0] / tt_size
            tr_ratio = self.train[self.train[self.group] == n].shape[0] / tr_size
            if tt_ratio >= tr_ratio:
                keep_tr += self.train[self.train[self.group] == n].index.tolist()
                continue

            ## 应抽样量
            tr_sample = int(tr_size * tt_ratio)

            ## 不重复抽样
            keep_tr += list(np.random.choice(self.train[self.train[self.group] == n].index,tr_sample,replace=False))
        ## 返回索引
        keep_tr = sorted(keep_tr)

        return self.train.loc[keep_tr,:]


def weighted_quantile(values, quantiles, sample_weight=None,
                      values_sorted=False, old_style=False):
    """ Very close to numpy.percentile, but supports weights.
    NOTE: quantiles should be in [0, 1]!
    :param values: numpy.array with data
    :param quantiles: array-like with many quantiles needed
    :param sample_weight: array-like of the same length as `array`
    :param values_sorted: bool, if True, then will avoid sorting of
        initial array
    :param old_style: if True, will correct output to be consistent
        with numpy.percentile.
    :return: numpy.array with computed quantiles.
    """
    values = np.array(values)
    quantiles = np.array(quantiles)
    if sample_weight is None:
        sample_weight = np.ones(len(values))
    sample_weight = np.array(sample_weight)
    assert np.all(quantiles >= 0) and np.all(quantiles <= 1), \
        'quantiles should be in [0, 1]'

    if not values_sorted:
        sorter = np.argsort(values)
        values = values[sorter]
        sample_weight = sample_weight[sorter]

    weighted_quantiles = np.cumsum(sample_weight) - 0.5 * sample_weight
    if old_style:
        # To be convenient with numpy.percentile
        weighted_quantiles -= weighted_quantiles[0]
        weighted_quantiles /= weighted_quantiles[-1]

    else:
        weighted_quantiles /= np.sum(sample_weight)
    return np.interp(quantiles, weighted_quantiles, values)

class Logger(object):
    level_relations = {
        'debug':logging.DEBUG,
        'info':logging.INFO,
        'warning':logging.WARNING,
        'error':logging.ERROR,
        'crit':logging.CRITICAL
    }#日志级别关系映射

    def __init__(self,filename,level='info',when='D',backCount=3,fmt='%(asctime)s - %(pathname)s[line:%(lineno)d] - %(levelname)s: %(message)s'):
        self.logger = logging.getLogger(filename)
        format_str = logging.Formatter(fmt)#设置日志格式
        self.logger.setLevel(self.level_relations.get(level))#设置日志级别
        sh = logging.StreamHandler()#往屏幕上输出
        sh.setFormatter(format_str) #设置屏幕上显示的格式
        th = handlers.TimedRotatingFileHandler(filename=filename,when=when,backupCount=backCount,encoding='utf-8')#往文件里写入#指定间隔时间自动生成文件的处理器
        #实例化TimedRotatingFileHandler
        #interval是时间间隔，backupCount是备份文件的个数，如果超过这个个数，就会自动删除，when是间隔的时间单位，单位有以下几种：
        # S 秒
        # M 分
        # H 小时、
        # D 天、
        # W 每星期（interval==0时代表星期一）
        # midnight 每天凌晨
        th.setFormatter(format_str)#设置文件里写入的格式
        self.logger.addHandler(sh) #把对象加到logger里
        self.logger.addHandler(th)
#     log = Logger('all.log',level='debug')
#     log.logger.debug('debug')
#     log.logger.info('info')
#     log.logger.warning('警告')
#     log.logger.error('报错')
#     log.logger.critical('严重')
#     Logger('error.log', level='error').logger.error('error')






import time
# from multiprocessing import Pool
from pathos.multiprocessing import Pool
import pandas as pd
import numpy as np

from lianyhaii.tools import time_it


def gen_date_feat(df_):
    df = df_.copy()
    df['hour'] = df['date'].dt.hour
    df['day'] = df['date'].dt.day
    df['weekday'] = df['date'].dt.weekday
    df['month'] = df['date'].dt.month
    df['holiday'] = df['weekday'].map(lambda x: 1 if x in [6,7] else 0)
    date_feat = ['hour','day','weekday','month','holiday']
    return df, date_feat

def add_lag_feat_byid(df_,col,d_shift=3):
    df = df_.copy()
    for diff in [1,3,12,24]:
        shift = d_shift + diff
        df[col+f"_shift_t{shift}"] = df.groupby(["MN_"])[col].transform(
            lambda x: x.shift(shift)
        )
    # for window in [3,12, 24]:
    #     df[col+f"_rolling_std_t{window}"] = df.groupby(["MN_"])[col].transform(
    #         lambda x: x.shift(d_shift).rolling(window).std()
    #     )

    for window in [3,6,12,24]:
        df[col+f"_rolling_mean_t{window}"] = df.groupby(["MN_"])[col].transform(
            lambda x: x.shift(d_shift).rolling(window).mean()
        )

    # for window in [3,6,12, 24]:
    #     df[col+f"_rolling_min_t{window}"] = df.groupby(["MN_"])[col].transform(
    #         lambda x: x.shift(d_shift).rolling(window).min()
    #     )
    # for window in [3,6,12,24]:
    #     df[col+f"_rolling_max_t{window}"] = df.groupby(["MN_"])[col].transform(
    #         lambda x: x.shift(d_shift).rolling(window).max()
    #     )

    # df[col+"_rolling_skew_t30"] = df.groupby(["MN_"])[col].transform(
    #     lambda x: x.shift(d_shift).rolling(30).skew()
    # )
    # df[col+"_rolling_kurt_t30"] = df.groupby(["MN_"])[col].transform(
    #     lambda x: x.shift(d_shift).rolling(30).kurt()
    # )

    ##���һ���ڵ�����վ���ͳ�����
    # for window in [1,3,6,24]:
    #     df[col+f"_rolling_mean_max_all_t{window}"]  = df.groupby['date'][col].mean().shift(d_shift).rolling(window).max()
    #     df[col+f"_rolling_max_max_all_t{window}"]  = df.groupby['date'][col].max().shift(d_shift).rolling(window).max()
    #
    # for window in [1,3,6,24]:
    #     df[col+f"_rolling_min_all_t{window}"]  = df[col].shift(d_shift).rolling(window).min()
    # for window in [1,3,6,24]:
    #     df[col+f"_rolling_mean_all_t{window}"]  = df[col].shift(d_shift).rolling(window).mean()
    #
    # for window in [1,3,6,24]:
    #     df[col+f"_rolling_std_all_t{window}"]  = df[col].shift(d_shift).rolling(window).std()
    # for diff in [1,3,6,24]:
    #     df[col+f"_shift_all_t{diff}"] = df.groupby('date')[col].mean().shift(d_shift+diff)

    ##����������

    for window in [1,3,12,24]:
        df[col+f"_rolling_ratio_t{window}"] = df.groupby('MN_')[col].transform(
            lambda x:x.pct_change(periods=window)
        )

    lag_feat = [x for x in df.columns if ('rolling' in x ) or ('shift' in x)]
    return df, lag_feat


def add_lag_feat(df_,col,d_shift=3):
    df = df_.copy()
    win_time = [3,5,7]
    more_time = [1] +win_time

    for diff in more_time:
        shift = d_shift + diff
        df[col+f"_shift_t{shift}"] = df[col].transform(
            lambda x: x.shift(shift)
        )
    # for window in win_time:
    #     df[col+f"_rolling_std_t{window}"] = df[col].transform(
    #         lambda x: x.shift(d_shift).rolling(window).std()
    #     )

    for window in win_time:
        df[col+f"_rolling_mean_t{window}"] = df[col].transform(
            lambda x: x.shift(d_shift).rolling(window).mean()
        )

    # for window in win_time:
    #     df[col+f"_rolling_min_t{window}"] = df[col].transform(
    #         lambda x: x.shift(d_shift).rolling(window).min()
    #     )
    # �����˷�λ��
    q_time = [5]
    # for window in q_time:
    #     df[col+f"_rolling_q70_t{window}"] = df[col].transform(
    #         lambda x: x.shift(d_shift).rolling(window).quantile(0.7)
    #     )
    # for window in q_time:
    #     df[col+f"_rolling_q30_t{window}"] = df[col].transform(
    #         lambda x: x.shift(d_shift).rolling(window).quantile(0.3)
    #     )


    ##����������
    for window in more_time:
        df[col+f"_rolling_ratio_t{window}"] = df[col].transform(
            lambda x:x.pct_change(periods=window)
        )

    lag_feat = [x for x in df.columns if ('rolling' in x ) or ('shift' in x)]

    return df, lag_feat


class ts_features():
    def __init__(self,train,test,ids,label,date_col,base_lag):
        self.train = train
        self.train['flat'] = 1
        self.test = test
        self.test['flat'] = 0
        self.ids = [ids] if type(ids) == str else ids
        self.label = label
        self.data = pd.concat([train,test],ignore_index=True,)
        self.date_col = date_col
        self.base_lag = base_lag
        # print(self.ids,[date_col])
        self.data.sort_values(
            by=self.ids+[date_col],
            ascending=True,
            inplace=True,
            ignore_index=True
        )

    @time_it
    def add_lag_feats(self,lags,isParallel=False):
        tt_feats = []
        ### 并行
        if isParallel:
            def lag_feats(df,l):
                return df.groupby(self.ids)[self.label].transform(lambda x: x.shift(l))
            res = []
            p = Pool()
            for lag in lags:
                lag += self.base_lag
                tmp = p.apply_async(lag_feats,args=(self.data[self.ids+[self.label]].copy(),lag),)
                res.append(tmp)
            p.close()
            p.join()
            for l,s in zip(lags,res):
                f_name = f'{self.label}_lag_{l}'
                # print(l,s)
                self.data[f_name] = s.get()
                tt_feats.append(f_name)

        if not isParallel:
            for l in lags:
                l += self.base_lag
                f_name = f'{self.label}_lag_{l}'
                self.data[f_name] = self.data.groupby(self.ids)[self.label].transform(lambda x:x.shift(l))
                # print(self.data)
                tt_feats.append(f_name)
        return tt_feats

    @time_it
    def add_rolling_feats(self,windows,method='mean'):
        tt_feats = []
        for window in windows:
            f_name = f'{self.label}_rolling_{method}_w{window}'
            if method =='mean':
                self.data[f_name] = self.data.groupby(self.ids)[self.label].transform(
                    lambda x: x.shift(self.base_lag).rolling(window).mean()
                )
            if method =='std':
                self.data[f_name] = self.data.groupby(self.ids)[self.label].transform(
                    lambda x: x.shift(self.base_lag).rolling(window).std()
                )
            tt_feats.append(f_name)

        return tt_feats

    def return_dataset(self):
        self.train = self.data[self.data['flat']==1]
        self.train.reset_index(drop=True,inplace=True)
        self.test = self.data[self.data['flat']==0]
        self.test.reset_index(drop=True,inplace=True)
        
        return self.train,self.test












import pandas as pd 
import numpy  as np 
import warnings

from lianyhaii.tools import transform_label 
warnings.filterwarnings('ignore')
import sys 
sys.path.append('../')
import lianyhaii

train = pd.DataFrame({
    'y':(np.random.random(size=500)>0.5).astype(int)
})
test = pd.DataFrame({
    'y':(np.random.random(size=500)>0.5).astype(int)
})
train['day'] = range(500)
test['day'] = range(500,1000)
for i in range(10):
    train[f'x{i}'] = np.random.random(size=500)
    test[f'x{i}'] = np.random.random(size=500)
label = 'y'

print(train.head())
lgb_params = {
    'objective': 'binary',
    'boosting_type': 'gbdt',
    'metric': 'auc',
    'n_jobs': 4,
    'verbose':-1,
}
from lianyhaii.ts_model import *
### todo
transform_type = 'ln'
label, fitted_lambda = transform_label(train, label, transform_type=transform_type)

mt = ts_test(train,test,base_features=[f'x{i}' for i in range(10)],new_features=[],
            m_score=[[0.90913,0.5]],label=label,metrices=['auc',],log_tool=None)
# mt.init_CV(seed=412,CV_type='skFold',n_split=5)
# oof,pred = mt.lgb_test(lgb_params=lgb_params)


import setuptools
import lianyhaii

with open("README.md", "r",encoding='utf-8') as fh:
    long_description = fh.read()

setuptools.setup(
    name="lianyhaii",
    version=f"{lianyhaii.__version__}",
    author="Ernnnn",
    author_email="lianyhai@163.com",
    description="A package to win data competition",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/Ernnnn/lianyhaii",
    packages=setuptools.find_packages(exclude=['build', 'dist', 'example', 'lianyhaii.egg-info']),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    install_requires=[
        'scikit-learn',
        'optuna',
        'plotly',
        'catboost',
        'xgboost',
        'lightgbm',
        'seaborn',
        'pyarrow',
        # 'missingno',
        # 'statsmodels',
        # 'pmdarima',
        # 'pytest',
        'gensim',
        # 'gokinjo',
        'pathos',
    ],
    include_package_data=True,
    zip_safe=False
)
#python setup.py sdist bdist_wheel
#python -m twine upload dist/*



# coding:utf-8
import os
import pickle
import time
from multiprocessing import Pool

import optuna
import plotly
from catboost import CatBoostClassifier
from lightgbm import LGBMClassifier,LGBMRegressor
import lightgbm as lgb
from optuna.visualization import plot_param_importances, plot_optimization_history
from sklearn import metrics
from sklearn.metrics import f1_score, confusion_matrix, accuracy_score, precision_recall_curve, auc,precision_score
from sklearn.model_selection import StratifiedKFold, KFold,GroupKFold
import seaborn as sns
import matplotlib.pyplot as plt
import pandas as pd
import numpy as np
# import statsmodels.api as sm
from sklearn.model_selection import train_test_split
from sklearn.metrics import roc_auc_score,precision_score,recall_score
from sklearn.linear_model import LinearRegression
# import pmdarima as pm
# from statsmodels.gam.tests.test_gam import sigmoid
from tqdm import tqdm
import xgboost as xgb

from lianyhaii.metrics import pymetric
from scipy.misc import derivative

def pr_auc(y_true, probas_pred):
    p, r, _ = precision_recall_curve(y_true, probas_pred)
    return auc(r, p)

def f2_score(y_true,y_pred):
    # print(y_true.shape)
    # print(y_pred.shape)
    # y_true = np.array(y_true)
    # y_pred = np.array(y_pred)
    # print(y_true.shape,y_pred.shape)
    # TP = np.sum((y_true@y_pred))
    # FP = np.sum(((~y_true)@y_pred))
    # FN = np.sum(((y_true)@(~y_pred)))
    # precision = TP/(TP+FP)
    # recall = TP/(TP+FN)
    recall = recall_score(y_true=y_true,y_pred=y_pred)
    precision = precision_score(y_true=y_true,y_pred=y_pred)
    F2 = 5*recall*precision/(4*precision+recall)
    return F2 



def tpr_weight_funtion(y_true,y_predict):
    d = pd.DataFrame()

    d['prob'] = list(y_predict)

    d['y'] = list(y_true)

    d = d.sort_values(['prob'], ascending=[0])

    y = d.y

    PosAll = pd.Series(y).value_counts()[1]

    NegAll = pd.Series(y).value_counts()[0]

    pCumsum = d['y'].cumsum()

    nCumsum = np.arange(len(y)) - pCumsum + 1

    pCumsumPer = pCumsum / PosAll

    nCumsumPer = nCumsum / NegAll

    TR1 = pCumsumPer[abs(nCumsumPer-0.001).idxmin()]

    TR2 = pCumsumPer[abs(nCumsumPer-0.005).idxmin()]

    TR3 = pCumsumPer[abs(nCumsumPer-0.01).idxmin()]
    # print(f'0.1% tpr is {TR1}, 0.5% tpr is {TR2} ,1% tpr is {TR3}')
    return 0.4 * TR1 + 0.3 * TR2 + 0.3 * TR3


def tpr_weight_cunstom(y_hat, data):
    y_true = data.get_label()
    y_hat = np.round(y_hat) # scikits f1 doesn't like probabilities
    return 'tpr_weight',tpr_weight_funtion(y_predict=y_hat,y_true=y_true),True

# def init_score(y_true):
#     # 样本初始值寻找过程
#     res = optimize.minimize_scalar(
#         lambda p: (y_true, p).sum(),
#         bounds=(0, 1),
#         method='bounded'
#     )
#     p = res.x
#     log_odds = np.log(p / (1 - p))
#     return log_odds

def focal_loss_lgb(y_pred, dtrain, alpha, gamma, num_class):
    """
    Focal Loss for lightgbm

    Parameters:
    -----------
    y_pred: numpy.ndarray
        array with the predictions
    dtrain: lightgbm.Dataset
    alpha, gamma: float
        See original paper https://arxiv.org/pdf/1708.02002.pdf
    num_class: int
        number of classes
    """
    a,g = alpha, gamma
    y_true = dtrain.label
    # N observations x num_class arrays
    y_true = np.eye(num_class)[y_true.astype('int')]
    y_pred = y_pred.reshape(-1,num_class, order='F')
    # alpha and gamma multiplicative factors with BCEWithLogitsLoss
    def fl(x,t):
        p = 1/(1+np.exp(-x))
        return -( a*t + (1-a)*(1-t) ) * (( 1 - ( t*p + (1-t)*(1-p)) )**g) * ( t*np.log(p)+(1-t)*np.log(1-p) )
    partial_fl = lambda x: fl(x, y_true)
    grad = derivative(partial_fl, y_pred, n=1, dx=1e-6)
    hess = derivative(partial_fl, y_pred, n=2, dx=1e-6)
    # flatten in column-major (Fortran-style) order
    return grad.flatten('F'), hess.flatten('F')

def focal_loss_lgb_eval_error(y_pred, dtrain, alpha, gamma, num_class):
    """
    Focal Loss for lightgbm

    Parameters:
    -----------
    y_pred: numpy.ndarray
        array with the predictions
    dtrain: lightgbm.Dataset
    alpha, gamma: float
        See original paper https://arxiv.org/pdf/1708.02002.pdf
    num_class: int
        number of classes
    """
    a,g = alpha, gamma
    y_true = dtrain.label
    y_true = np.eye(num_class)[y_true.astype('int')]
    y_pred = y_pred.reshape(-1, num_class, order='F')
    p = 1/(1+np.exp(-y_pred))
    loss = -( a*y_true + (1-a)*(1-y_true) ) * (( 1 - ( y_true*p + (1-y_true)*(1-p)) )**g) * ( y_true*np.log(p)+(1-y_true)*np.log(1-p) )
    # a variant can be np.sum(loss)/num_class
    return 'focal_loss', np.mean(loss), False


def lgb_f1_score(y_hat, data):
    y_true = data.get_label()
    y_hat = np.round(y_hat) # scikits f1 doesn't like probabilities
    is_higher_better = True
    return 'f1', f1_score(y_true=y_true, y_pred=y_hat,average='macro'), is_higher_better


def mape(y_true, y_pred):
    return np.mean(np.abs((y_pred - y_true) / y_true)) * 100

def smape(y_true, y_pred):
    return 2.0 * np.mean(np.abs(y_pred - y_true) / (np.abs(y_pred) + np.abs(y_true))) * 100

def f1_score_custom(y_true, y_pred):
    y_pred = y_pred.round()
    return 'f1', f1_score(y_true, y_pred), True





class make_test():
    def __init__(self,tr_df,tt_df,base_features,new_features,m_score,label,metrices=None,log_tool=None):
        print(tr_df.shape,tt_df.shape)
        print(base_features+new_features)
        self.train = tr_df
        self.test = tt_df
        self.base_features = base_features
        self.new_features = new_features
        self.m_score = m_score
        self.label = label
        self.features = base_features + new_features
        self.predictions = None
        self.model = []
        self.features_imp = []
        self.metrices = metrices
        self.run = log_tool


    def init_CV(self,seed,n_split=5,shuffle=True,CV_type='skFold',group_col=None):
        self.cv_conf = {}
        if  self.run is not None:
            self.run['cv_type'] = CV_type
        if CV_type == 'skFold':
            cv = StratifiedKFold(n_splits=n_split,shuffle=shuffle,random_state=seed)
            self.cv_conf['iter'] = cv.split(self.train[self.features],self.train[self.label])
            self.cv_conf['n'] = n_split

        elif CV_type == 'kFold':
            cv = KFold(n_splits=n_split, shuffle=shuffle, random_state=seed)
            self.cv_conf['iter'] = [x for x in cv.split(X=self.train[self.features], y=self.train[self.label])]

            self.cv_conf['n'] = n_split

        elif CV_type == 'lastFold':
            folds = sorted(self.train[group_col].unique())

            cv = [[self.train[self.train[group_col]<folds[-1]].index,
                 self.train[(self.train[group_col]==folds[-1])].index]]
            self.cv_conf['iter'] = cv
            self.cv_conf['n'] = 1

        elif CV_type == 'gFold':
            cv = GroupKFold(n_splits=n_split,)
            self.cv_conf['iter'] = cv.split(self.train[self.features],y=self.train[self.label],groups=self.train[group_col])
            self.cv_conf['n'] = n_split
        elif CV_type == 'online':
            folds = sorted(self.train[group_col].unique())
            cv = [[self.train.index,self.train[(self.train[group_col]==folds[-1])].index]]

            self.cv_conf['iter'] = cv
            self.cv_conf['n'] = 1

        else:
            raise ValueError('no this type of fold')

    def __deal_cat_features(self,cat_features,m_type='lgb'):
        from sklearn.preprocessing import LabelEncoder
        if m_type == 'lgb':
            for col in cat_features:
                if self.train[col].dtype.name != 'category':
                    self.train[col] = self.train[col].fillna('unseen_before').astype(str)
                    self.test[col] = self.test[col].fillna('unseen_before').astype(str)


                    le = LabelEncoder()
                    le.fit(list(self.train[col])+list(self.test[col]))
                    self.train[col] = le.transform(self.train[col])
                    self.test[col] = le.transform(self.test[col])

                    self.train[col] = self.train[col].astype('category')
                    self.test[col] = self.test[col].astype('category')
    def __check_diff_score(self,oof_predictions,val_idx=None):

        if val_idx is None:
            pm = pymetric(y_true=self.train[self.label],y_pred=oof_predictions)
        else:
            pm = pymetric(y_true=self.train.loc[val_idx,self.label],y_pred=oof_predictions[val_idx])
        ## 加入新函数
        old_metrics = [x for x in self.metrices if type(x) == str]
        new_metrics = [x for x in self.metrices if type(x) != str]
        for name,f in new_metrics:
            pm.add_metric(name,f)
        old_metrics += [x for x,_ in new_metrics]
        result_score = pm.gen_metric_dict(metric_names=old_metrics,th=0.5)
        
        for key,value in result_score.items():
            print(f'global {key} : {value}')
            if  self.run is not None :
                self.run[f'metrics/global_{key}'] = value
        print('='*10+'different with previous version'+'='*10)
        score_list = []
        for n,(key,value) in enumerate(result_score.items()):
            print(f'diff of {key} : {np.round(value-self.m_score[-1][n],5)}')
            if (key == 'auc')&(self.run is not None)*(np.round(value-self.m_score[-1][n],5)>0):
                self.run['sys/tags'].add(['boosted'])
            score_list.append(value)

        self.m_score.append(score_list)

    def lgb_test(self,lgb_params,cv_score=False,weight=None):
        # self.__deal_cat_features(cat_features,m_type='lgb')
        # imp = pd.DataFrame({
        #     'feature':self.features,
        #     'gain':0,
        #     'split':0,
        # })

        feature_imp = np.zeros(len(self.features))
        cv_score_list = []
        oof_predictions = np.zeros(len(self.train))
        tt_predicts = np.zeros(len(self.test))

        for n,(trn,val) in enumerate(self.cv_conf['iter']):
            trn_X,trn_y = self.train.loc[trn,self.features],self.train.loc[trn,self.label]
            val_X,val_y = self.train.loc[val,self.features],self.train.loc[val,self.label]
            if weight is not None:
                trn_data = lgb.Dataset(trn_X,label=trn_y,weight=self.train.loc[trn,weight])
            else:
                trn_data = lgb.Dataset(trn_X,label=trn_y)
            # trn_data = lgb.Dataset(trn_X,label=trn_y)
            val_data = lgb.Dataset(val_X,label=val_y)

            estimator = lgb.train(lgb_params,
                                  trn_data,
                                  valid_sets=[trn_data,val_data],
                                  # fobj=f1_macro_loss,
                                  # feval=lgb_f1_score,
                                  # feval=tpr_weight_3_cunstom,
                                  verbose_eval=-1,
                                  )

            oof_predictions[val] = estimator.predict(val_X)
            self.model.append(estimator)
            cv_score_list.append(roc_auc_score(y_true=val_y,y_score=oof_predictions[val]))
            tt_predicts += estimator.predict(self.test[self.features]) / self.cv_conf['n']
            # imp['gain'] = estimator.feature_importance(importance_type='gain') / self.cv_conf['n']
            feature_imp += estimator.feature_importance(importance_type='split') / self.cv_conf['n']
            if self.run is not None:
                self.run['metrics/test_auc'].log(cv_score_list[-1])

        print(f"training CV oof mean : {np.round(np.mean(cv_score_list),5)}")
        if self.cv_conf['n']==1:
            self.__check_diff_score(oof_predictions,val_idx=val)
        else:
            self.__check_diff_score(oof_predictions)
        self.predictions = tt_predicts
        self.features_imp = feature_imp
        if self.run is not None:
            self.run['feature_importance'] = self.features_importance()
        if cv_score:
            return oof_predictions,tt_predicts,cv_score_list
        else:
            if self.cv_conf['n']==1:
                # print(oof_predictions.shape,oof_predictions[val].shape,)
                return oof_predictions[val],tt_predicts,val_y.values

            return oof_predictions,tt_predicts

    def xgb_test(self,xgb_params,cv_score=False):
        cv_score_list = []
        oof_predictions = np.zeros(len(self.train))
        tt_predicts = np.zeros(len(self.test))

        for n,(trn,val) in enumerate(self.cv_conf['iter']):
            trn_X,trn_y = self.train.loc[trn,self.features],self.train.loc[trn,self.label]
            val_X,val_y = self.train.loc[val,self.features],self.train.loc[val,self.label]

            tr_data = xgb.DMatrix(trn_X, label=trn_y)
            val_X = xgb.DMatrix(val_X,label=val_y)

            watchlist = [(tr_data, 'train'), (val_X, 'eval')]
            estimator = xgb.train(
                xgb_params,
                tr_data,
                evals=watchlist,
                verbose_eval=100,
                num_boost_round=30000,
                # early_stopping_rounds=200,
                early_stopping_rounds=100,
            )

            oof_predictions[val] = estimator.predict(val_X)

            cv_score_list.append(roc_auc_score(y_true=val_y,y_score=estimator.predict(val_X)))
            tt_predicts += estimator.predict(xgb.DMatrix(self.test[self.features])) / self.cv_conf['n']
            if self.run is not None:
                self.run['metrics/test_auc'].log(cv_score_list[-1])

        self.__check_diff_score(oof_predictions)
        # print(imp.sort_values(['split','gain'],ascending=False,ignore_index=True).loc[imp['feature'].isin(self.new_features),:])
        self.predictions = tt_predicts
        if cv_score:
            if self.cv_conf['n']==1:
                # print(oof_predictions.shape,oof_predictions[val].shape,)
                return oof_predictions[val],tt_predicts,val_y.values

            return oof_predictions,tt_predicts

    def cat_test(self,cat_params,cv_score=False,cat_features=None,save_path='.'):
        cv_score_list = []
        oof_predictions = np.zeros(len(self.train))
        tt_predicts = np.zeros(len(self.test))
        feature_imp = np.zeros(len(self.features))
        if cat_features is None:
            cat_features = [x for x in self.train.select_dtypes(include=['category']).columns.tolist() if x in self.features]

        for c in cat_features:
            self.train[c] = self.train[c].astype('str')
            self.test[c] = self.test[c].astype('str')

        for n,(trn,val) in enumerate(self.cv_conf['iter']):
            trn_X,trn_y = self.train.loc[trn,self.features],self.train.loc[trn,self.label]
            val_X,val_y = self.train.loc[val,self.features],self.train.loc[val,self.label]

            estimator = CatBoostClassifier(**cat_params)
            print(cat_params)
            print(cat_params)
            if os.path.exists(save_path+f'/model_cbt_f{n}.cbt'):
                estimator.load_model(save_path+f'/model_cbt_f{n}.cbt')
            else:
                estimator.fit(
                    trn_X,trn_y,
                    cat_features=cat_features,
                    # early_stopping_rounds=100,
                    # eval_set=[(val_X,val_y)],
                    eval_set=[(val_X,val_y)] if cat_params['task_type'] == 'GPU' else [(trn_X,trn_y),(val_X,val_y)],
                    use_best_model=True,
                    metric_period=200,
                    verbose=True,
                )
                estimator.save_model(save_path+f'/model_cbt_f{n}.cbt')
            oof_predictions[val] = estimator.predict_proba(val_X)[:,1]
            self.model.append(estimator)
            cv_score_list.append(roc_auc_score(y_true=val_y,y_score=estimator.predict_proba(val_X)[:,1]))
            tt_predicts += estimator.predict_proba(self.test[self.features])[:,1] / self.cv_conf['n']
            feature_imp += estimator.feature_importances_ / self.cv_conf['n']
            if self.run is not None:
                self.run['metrics/test_auc'].log(cv_score_list[-1])

        print(f"training CV oof mean : {np.round(np.mean(cv_score_list), 5)}")
        self.__check_diff_score(oof_predictions,val_idx=val)
        self.predictions = tt_predicts
        self.features_imp = feature_imp
        if self.run is not None:
            self.run['feature_importance'] = self.features_importance()
        if cv_score:
            return oof_predictions,tt_predicts,cv_score_list
        else:
            if self.cv_conf['n']==1:
                # print(oof_predictions.shape,oof_predictions[val].shape,)
                return oof_predictions[val],tt_predicts,val_y.values

            return oof_predictions,tt_predicts

    def sklearn_test(self, sklean_model, cv_score=False, weight=None):
        # self.__deal_cat_features(cat_features,m_type='lgb')
        # imp = pd.DataFrame({
        #     'feature':self.features,
        #     'gain':0,
        #     'split':0,
        # })

        feature_imp = np.zeros(len(self.features))
        cv_score_list = []
        oof_predictions = np.zeros(len(self.train))
        tt_predicts = np.zeros(len(self.test))

        for n, (trn, val) in enumerate(self.cv_conf['iter']):
            trn_X, trn_y = self.train.loc[trn, self.features], self.train.loc[trn, self.label]
            val_X, val_y = self.train.loc[val, self.features], self.train.loc[val, self.label]


            estimator = sklean_model.fit(trn_X,trn_y)

            oof_predictions[val] = estimator.predict_proba(val_X)[:,1]
            self.model.append(estimator)

            cv_score_list.append(roc_auc_score(y_true=val_y, y_score=oof_predictions[val]))
            tt_predicts += estimator.predict_proba(self.test[self.features])[:,1] / self.cv_conf['n']
            
            if self.run is not None:
                self.run['metrics/test_auc'].log(cv_score_list[-1])


        print(f"training CV oof mean : {np.round(np.mean(cv_score_list), 5)}")

        self.__check_diff_score(oof_predictions)
        self.predictions = tt_predicts
        self.features_imp = feature_imp

        if self.run is not None:
            self.run['feature_importance'] = self.features_importance()
        if cv_score:
            return oof_predictions,tt_predicts,cv_score_list
        else:
            if self.cv_conf['n']==1:
                # print(oof_predictions.shape,oof_predictions[val].shape,)
                return oof_predictions[val],tt_predicts,val_y.values

            return oof_predictions,tt_predicts

    def find_threshold(self,oof,ylabel,weight_range,method='offline',predictions=None,online_thd=None):
        if method == 'offline':
            res = pd.DataFrame({
                'weight': weight_range,
                'oof_f1': 0,
                # 'oof_auc': 0,
            })
            # def record_f1(th):
            #     oof_f1 = f1_score(y_true=ylabel, y_pred=[int(x > n) for x in oof])
            #     # oof_auc = roc_auc_score(y_score=oof,y_true=ylabel)
            #
            #     # print('the weighted oof auc : ',oof_auc)
            #     print(f'the weighted {np.round(n, 4)} oof f1 {np.round(oof_f1, 4)} ')
            #     # res.loc[res['weight'] == n,'oof_auc'] = oof_auc
            #     res.loc[res['weight'] == n, 'oof_f1'] = oof_f1
            # p= Pool()
            # p.apply_async(record_f1,args=weight_range,)
            # p.join()
            # p.close()


            for n in weight_range:
                # oof_f1 = f1_score(y_true=ylabel,y_pred=[int(x>n) for x in oof ])
                # print(ylabel.shape,oof.shape)
                oof_f1 = f2_score(y_true=ylabel,y_pred=np.array([int(x>n) for x in oof]))
                # oof_auc = roc_auc_score(y_score=oof,y_true=ylabel)

                # print('the weighted oof auc : ',oof_auc)
                print(f'the weighted {np.round(n,4)} oof f1 {np.round(oof_f1,4)} ')
                # res.loc[res['weight'] == n,'oof_auc'] = oof_auc
                res.loc[res['weight'] == n,'oof_f1'] = oof_f1
            fig = plt.figure(figsize=(8, 6))
            ax2 = fig.add_subplot(111)
            res.index = res['weight']

            res['oof_f1'].plot(ax=ax2, style='y>-.', alpha=0.7, legend=True)
            plt.legend(loc=1)
            plt.show()
            best_threshold = res.loc[res['oof_f1']==res['oof_f1'].max(),'weight'].values
            print(f'the best threshold {best_threshold} and best f1 {res["oof_f1"].max()}')

        elif method == 'ythreshold':

            print(f'train dataset 01 distributions :{sum(ylabel)/len(ylabel)}')
            preds_index = int(len(predictions)*(1-(sum(ylabel)/len(ylabel))))
            sorted_pred = sorted(predictions)
            best_threshold = sorted_pred[preds_index]

            ##
            oof_index = sum(ylabel)
            sorted_oof = sorted(oof)
            best_oof = sorted_oof[oof_index]

            res = f1_score(y_true=ylabel,y_pred=[int(x>=best_oof) for x in oof],average='macro')
            print(f'weight is {best_oof} and test weight {best_threshold} oof was {res}')
        elif method == 'online':
            preds_index = int(len(predictions)*(online_thd/(1-online_thd)))
            sorted_pred = sorted(predictions)
            best_threshold = sorted_pred[preds_index]
            print(f'online wight is {best_threshold}')
            res = 0

        elif method == 'sortedPvalue':
            best_threshold,res = 0,0
            predictions = sorted(list(predictions))

            plt.figure(figsize=(20,10))
            sns.lineplot(predictions,)
            plt.show()
        else:
            raise NotImplementedError

        return best_threshold,res

    def features_importance(self):
        df = pd.DataFrame(self.features_imp, index=self.features,columns=['values'])
        df.sort_values('values',ascending=False,inplace=True,)

        return df

    def analysis_residual(self,oof_predictions):
        self.train['predic_res'] = oof_predictions
        res_train_t = self.train[self.train[self.label] == np.round(oof_predictions)].copy()
        res_train_f = self.train[self.train[self.label] != np.round(oof_predictions)].copy()
        try:
            import os
            os.mkdir('res_analysis')
        except:
            pass
        today = time.strftime("%Y-%m-%d", time.localtime())[5:]
        res_train_f.to_csv(f'./res_analysis/res_train_f_{today}.csv',index=False)
        res_train_t.to_csv(f'./res_analysis/res_train_t_{today}.csv',index=False)

    def submit_seeds_ensembel(self,lgb_params,seed_list):
        """
        the aim of fuction is to find good seed or ensembel seek to submit.

        """
        tt_predictions = np.zeros(len(self.test))
        seed_df = pd.DataFrame({
            'seed':seed_list,
            'oof_score':0,
            'cv_mean':0,
            'cv_std':0,
        })
        final_oof = np.zeros(len(self.train))

        for n,s in enumerate(seed_list):
            # seed_everything(s)
            self.init_CV(seed=s)
            oof_preds,tt_preds,cv_list = self.lgb_test(lgb_params=lgb_params,cv_score=True)
            tt_predictions += tt_preds/len(seed_list)
            final_oof += oof_preds/len(seed_df)

            seed_df.loc[n,'oof_auc'] = roc_auc_score(y_score=oof_preds,y_true=self.train[self.label])
            seed_df.loc[n,'oof_score'] = tpr_weight_funtion(y_predict=oof_preds,y_true=self.train[self.label])
            seed_df.loc[n,'cv_mean'] = np.mean(cv_list)
            seed_df.loc[n,'cv_std'] = np.std(cv_list)
            print(seed_df.loc[n,:],end='\n')

        self.predictions = tt_predictions
        global_auc = roc_auc_score(y_true=self.train[self.label],y_score=final_oof)
        global_tpr = tpr_weight_funtion(y_true=self.train[self.label],y_predict=final_oof)
        self.m_score.append([global_auc,global_tpr])
        seed_df['diff_abs'] = (seed_df['oof_auc'] - seed_df['cv_mean']).abs()
        seed_df.sort_values(['diff_abs','cv_std'],inplace=True)
        return seed_df

    def submit_pseudo_label_model(self,params,seed,model_type):
        # if len(self.predictions) == 0:
        print('training pseudo label .....')
        self.test['label'] = self.predictions.copy()
        test2p = self.test[(self.test['label']<=0.01)|(self.test['label']>=0.95)].copy()
        # relabel
        test2p.loc[test2p['label'] >= 0.5,'label'] = 1
        test2p.loc[test2p['label'] < 0.5,'label'] = 0

        train2p = pd.concat([self.train,test2p],axis=0)
        train2p.reset_index(drop=True,inplace=True)

        tmp_mt = make_test(train2p,self.test,self.features,[],m_score=self.m_score,label=self.label)
        tmp_mt.init_CV(seed)
        if model_type =='lgb':
            tmp_mt.lgb_test(params)
        if model_type == 'xgb':
            tmp_mt.xgb_test(params)

        self.predictions = tmp_mt.predictions

    def submit(self,ID,sub_file=True,threshold=None):
        today = time.strftime("%Y-%m-%d", time.localtime())[5:]
        # self.test[self.label] = [int(x) for x in self.predictions > 0.5]
        if threshold is None:
            self.test[self.label] = self.predictions
        else:
            self.test[self.label] = [int(x >= threshold) for x in self.predictions]
            print('sum of label',np.sum(self.test[self.label]))

        sub_train = self.test[[ID,self.label]].copy()
        if sub_file:
            sub_test = pd.read_csv('./data/submit.csv')
            sub = sub_test[[ID]].merge(sub_train, on=ID, how='left')
        else:
            sub = sub_train.copy()

        # plt.figure(figsize=(20, 10))
        # sns.distplot(sub['label'], bins=100)
        # plt.show()
        print('null in sub','\n',sub.isnull().sum())
        sub.fillna(0, inplace=True)
        score = str(np.round(self.m_score[-1][0],4))+"_"+str(np.round(self.m_score[-1][1],4))
        sub.to_csv(f'./result/sub_{today}_{score}.csv', index=False)
        return sub

    def psm_samples(self,lgb_params):

        ## p值
        tr,tt = self.train.copy(),self.test.copy()
        tr['test_label'] = 0
        tt['test_label'] = 1
        data = pd.concat([tr,tt],ignore_index=True,)
        # train,test = train_test_split(data,test_size=0.1,random_state=0)

        mt = make_test(data,data,self.features,[],m_score=[[0,0,0]],label='test_label')
        mt.init_CV(0)
        oof , test = mt.lgb_test(lgb_params)
        data['oof'] = oof
        ## 可视化
        # tmp_df = self.train[[self.label,'oof']]
        sns.boxplot(x='label',y='oof',data=data)
        plt.show()
        ## 找特征重要性
        imp_df = mt.features_importance()
        imp_df.reset_index(drop=False,inplace=True)
        imp_df.columns = ['features','values']
        sns.barplot(y='features',x='values',data=imp_df)
        plt.show()
        print(imp_df.head(10))
        return data,imp_df

    def submit_easy_ensemble(self,n_group=10,):
        pass

    def tune_model(self,params,model,type='binary',N_TRIALS=100,model_type='lgb'):

        def objective(trial):
            if model_type == 'lgb':
                obj_params = {
                    # 'random_state': params['seed'],
                    'metric': params['metric'],
                    'n_estimators': 5000,
                    'n_jobs': -1,
                    'seed':params['seed'],
                    'early_stopping_rounds': 100,
                    'learning_rate': trial.suggest_float('learning_rate', 1e-4, 0.1),
                    'max_depth': trial.suggest_int('max_depth', 6, 16),
                    'num_leaves': trial.suggest_int('num_leaves', 31, 2**8),
                    'reg_alpha': trial.suggest_float('reg_alpha', 1e-3, 10.0),
                    'reg_lambda': trial.suggest_float('reg_lambda', 1e-3, 10.0),
                    'colsample_bytree': trial.suggest_float('colsample_bytree', 0.2, 0.9),
                    'min_child_samples': trial.suggest_int('min_child_samples', 1, 300),
                    'subsample_freq': trial.suggest_int('subsample_freq', 1, 10),
                    'subsample': trial.suggest_float('subsample', 0.3, 0.9),
                    'max_bin': trial.suggest_int('max_bin', 128, 1024),
                    'min_data_per_group': trial.suggest_int('min_data_per_group', 50, 200),
                    'cat_smooth': trial.suggest_int('cat_smooth', 10, 100),
                    'cat_l2': trial.suggest_int('cat_l2', 1, 20),
                    'verbose': -1,
                      }

            elif model_type == 'xgb':
                obj_params = {
                    'objective': params['objective'],
                    # 'task_type': 'GPU',
                    'nthread': -1,
                    'seed': params['seed'],
                    'tree_method': 'hist',
                    'eval_metric': params['eval_metric'],
                    'lambda': trial.suggest_loguniform('lambda', 1e-3, 10.0),
                    'alpha': trial.suggest_loguniform('alpha', 1e-3, 10.0),
                    'colsample_bytree': trial.suggest_categorical('colsample_bytree', [0.3,0.4,0.5,0.6,0.7,0.8,0.9, 1.0]),
                    'subsample': trial.suggest_categorical('subsample', [0.4,0.5,0.6,0.7,0.8,1.0]),
                    'learning_rate': trial.suggest_categorical('learning_rate', [0.008,0.009,0.01,0.012,0.014,0.016,0.018, 0.02]),
                    # 'n_estimators': 10000,
                    'max_depth': trial.suggest_int('max_depth', 5,14),
                    # 'min_child_weight': trial.suggest_int('min_child_weight', 1, 300),
                }

            else:
                raise NotImplementedError
            oof, _ = model(obj_params, cv_score=False,)
            score = 0
            if type == 'binary':
                score = roc_auc_score(y_true=self.train[self.label],y_score=oof)
            if type == 'f1':
                score = f1_score(y_true=self.train[self.label],y_pred=[int(x>0.5) for x in oof],average='macro')

            return score

        study = optuna.create_study(study_name=f"optimization", direction='maximize')
        study.optimize(objective, n_trials=N_TRIALS)
        print('Number of finished trials:', len(study.trials))
        print('Best trial:', study.best_trial.params)
        ## save log
        print(study.trials_dataframe())
        import os
        if not os.path.exists('user_data'):
            os.mkdir('user_data')

        study.trials_dataframe().to_csv('./user_data/' + f"{type}trial_parameters.csv", index=False)
        with open('./user_data/' + f'{type}_study.pickle', 'wb') as f:
            pickle.dump(study, f)

        fig = plot_param_importances(study,)
        plotly.offline.plot(fig, filename='imp_fig.html')
        fig = plot_optimization_history(study)
        plotly.offline.plot(fig, filename='history_fig.html')

        return study.best_trial.params



class multi_class_test():
    def __init__(self,tr_df,tt_df,num_class,base_features,new_features,m_score,label,metrices=None,log_tool=None):
        print(tr_df.shape,tt_df.shape)
        print(base_features+new_features)
        self.train = tr_df
        self.test = tt_df
        self.base_features = base_features
        self.new_features = new_features
        self.m_score = m_score
        self.label = label
        self.features = base_features + new_features
        self.predictions = None
        self.model = []
        self.features_imp = []
        self.metrices = metrices
        self.run = log_tool
        self.num_class = num_class


    def init_CV(self,seed,n_split=5,shuffle=True,CV_type='skFold',group_col=None):
        self.cv_conf = {}
        if  self.run is not None:
            self.run['cv_type'] = CV_type
        if CV_type == 'skFold':
            cv = StratifiedKFold(n_splits=n_split,shuffle=shuffle,random_state=seed)
            self.cv_conf['iter'] = cv.split(self.train[self.features],self.train[self.label])
            self.cv_conf['n'] = n_split

        elif CV_type == 'kFold':
            cv = KFold(n_splits=n_split, shuffle=shuffle, random_state=seed)
            self.cv_conf['iter'] = [x for x in cv.split(X=self.train[self.features], y=self.train[self.label])]

            self.cv_conf['n'] = n_split

        elif CV_type == 'lastFold':
            folds = sorted(self.train[group_col].unique())

            cv = [[self.train[self.train[group_col]<folds[-1]].index,
                 self.train[(self.train[group_col]==folds[-1])].index]]
            self.cv_conf['iter'] = cv
            self.cv_conf['n'] = 1

        elif CV_type == 'gFold':
            cv = GroupKFold(n_splits=n_split,)
            self.cv_conf['iter'] = cv.split(self.train[self.features],y=self.train[self.label],groups=self.train[group_col])
            self.cv_conf['n'] = n_split
        elif CV_type == 'online':
            folds = sorted(self.train[group_col].unique())
            cv = [[self.train.index,self.train[(self.train[group_col]==folds[-1])].index]]

            self.cv_conf['iter'] = cv
            self.cv_conf['n'] = 1

        else:
            raise ValueError('no this type of fold')

    def __deal_cat_features(self,cat_features,m_type='lgb'):
        from sklearn.preprocessing import LabelEncoder
        if m_type == 'lgb':
            for col in cat_features:
                if self.train[col].dtype.name != 'category':
                    self.train[col] = self.train[col].fillna('unseen_before').astype(str)
                    self.test[col] = self.test[col].fillna('unseen_before').astype(str)


                    le = LabelEncoder()
                    le.fit(list(self.train[col])+list(self.test[col]))
                    self.train[col] = le.transform(self.train[col])
                    self.test[col] = le.transform(self.test[col])

                    self.train[col] = self.train[col].astype('category')
                    self.test[col] = self.test[col].astype('category')
    def __check_diff_score(self,oof_predictions,val_idx=None):

        if val_idx is None:
            pm = pymetric(y_true=self.train[self.label],y_pred=oof_predictions)
        else:
            pm = pymetric(y_true=self.train.loc[val_idx,self.label],y_pred=oof_predictions[val_idx])

        result_score = pm.gen_metric_dict(metric_names=self.metrices,th=0.5)
        for key,value in result_score.items():
            print(f'global {key} : {value}')
            if  self.run is not None :
                self.run[f'metrics/global_{key}'] = value
        print('='*10+'different with previous version'+'='*10)
        score_list = []
        for n,(key,value) in enumerate(result_score.items()):
            print(f'diff of {key} : {np.round(value-self.m_score[-1][n],5)}')
            if (key == 'auc')&(self.run is not None)*(np.round(value-self.m_score[-1][n],5)>0):
                self.run['sys/tags'].add(['boosted'])
            score_list.append(value)

        self.m_score.append(score_list)

    def lgb_test(self,lgb_params,cv_score=False,weight=None,is_save=False):
        # self.__deal_cat_features(cat_features,m_type='lgb')
        # imp = pd.DataFrame({
        #     'feature':self.features,
        #     'gain':0,
        #     'split':0,
        # })

        feature_imp = np.zeros(len(self.features))
        cv_score_list = []
        oof_predictions = np.zeros((len(self.train),self.num_class))
        tt_predicts = np.zeros((len(self.test),self.num_class))

        for n,(trn,val) in enumerate(self.cv_conf['iter']):
            trn_X,trn_y = self.train.loc[trn,self.features],self.train.loc[trn,self.label]
            val_X,val_y = self.train.loc[val,self.features],self.train.loc[val,self.label]
            if weight is not None:
                trn_data = lgb.Dataset(trn_X,label=trn_y,weight=self.train.loc[trn,weight])
            else:
                trn_data = lgb.Dataset(trn_X,label=trn_y)
            # trn_data = lgb.Dataset(trn_X,label=trn_y)
            val_data = lgb.Dataset(val_X,label=val_y)

            estimator = lgb.train(lgb_params,
                                  trn_data,
                                  valid_sets=[trn_data,val_data],
                                #   fobj=lambda x,y: focal_loss_lgb(x, y, 0.25, 2., self.num_class),
                                #   feval=lambda x,y: focal_loss_lgb_eval_error(x, y, 0.25, 2., self.num_class),
                                  # feval=tpr_weight_3_cunstom,
                                  verbose_eval=-1,
                                  )

            oof_predictions[val,:] = estimator.predict(val_X)
            self.model.append(estimator)
            
            # cv_score_list.append(roc_auc_score(y_true=val_y,y_score=oof_predictions[val]))
            cur_oof = np.argmax(oof_predictions[val,:],axis=1)
            print(oof_predictions[val,:].shape)
            # print()
            print(cur_oof.shape)
            acc = np.sum(cur_oof == np.array(val_y))/(len(val_y))
            cv_score_list.append(acc)
            print('cur acc',acc)
            tt_predicts += estimator.predict(self.test[self.features]) / self.cv_conf['n']
            # imp['gain'] = estimator.feature_importance(importance_type='gain') / self.cv_conf['n']
            feature_imp += estimator.feature_importance(importance_type='split') / self.cv_conf['n']
            if self.run is not None:
                self.run['metrics/test_auc'].log(cv_score_list[-1])

        print(f"training CV oof mean : {np.round(np.mean(cv_score_list),5)}")
        cur_oof = np.argmax(oof_predictions,axis=1)
        print(cur_oof.shape)
        y_label = self.train[self.label].values
        acc = np.sum(cur_oof == np.array(y_label))/(len(y_label))
        print('global acc ',acc)
        # self.__check_diff_score(oof_predictions,val_idx=val)
        self.predictions = tt_predicts
        self.features_imp = feature_imp
        if self.run is not None:
            self.run['feature_importance'] = self.features_importance()
        if cv_score:
            return oof_predictions,tt_predicts,cv_score_list
        else:
            if self.cv_conf['n']==1:
                # print(oof_predictions.shape,oof_predictions[val].shape,)
                return oof_predictions[val],tt_predicts,val_y.values

            return oof_predictions,tt_predicts

    def xgb_test(self,xgb_params,cv_score=False):
        cv_score_list = []
        oof_predictions = np.zeros(len(self.train))
        tt_predicts = np.zeros(len(self.test))

        for n,(trn,val) in enumerate(self.cv_conf['iter']):
            trn_X,trn_y = self.train.loc[trn,self.features],self.train.loc[trn,self.label]
            val_X,val_y = self.train.loc[val,self.features],self.train.loc[val,self.label]

            tr_data = xgb.DMatrix(trn_X, label=trn_y)
            val_X = xgb.DMatrix(val_X,label=val_y)

            watchlist = [(tr_data, 'train'), (val_X, 'eval')]
            estimator = xgb.train(
                xgb_params,
                tr_data,
                evals=watchlist,
                verbose_eval=100,
                num_boost_round=30000,
                # early_stopping_rounds=200,
                early_stopping_rounds=100,
            )

            oof_predictions[val] = estimator.predict(val_X)

            cv_score_list.append(roc_auc_score(y_true=val_y,y_score=estimator.predict(val_X)))
            tt_predicts += estimator.predict(xgb.DMatrix(self.test[self.features])) / self.cv_conf['n']
            if self.run is not None:
                self.run['metrics/test_auc'].log(cv_score_list[-1])

        print(f"training CV oof mean : {np.round(np.mean(cv_score_list),5)}")
        if self.cv_conf['n']==1:
            self.__check_diff_score(oof_predictions,val_idx=val)
        else:
            self.__check_diff_score(oof_predictions)
        # print(imp.sort_values(['split','gain'],ascending=False,ignore_index=True).loc[imp['feature'].isin(self.new_features),:])
        self.predictions = tt_predicts
        if cv_score:
            return oof_predictions, tt_predicts, cv_score_list
        else:
            # if self.cv_conf['n'] == 1:
            return oof_predictions[val],tt_predicts,val_y
            # else:
                # return oof_predictions, tt_predicts

    def cat_test(self,cat_params,cv_score=False,cat_features=None,save_path:str='.'):
        """
        save_path:"/home/lianyhaii"
        """
        cv_score_list = []
        oof_predictions = np.zeros(len(self.train))
        tt_predicts = np.zeros(len(self.test))
        if cat_features is None:
            cat_features = [x for x in self.train.select_dtypes(include=['category']).columns.tolist() if x in self.features]

        for c in cat_features:
            self.train[c] = self.train[c].astype('str')
            self.test[c] = self.test[c].astype('str')

        for n,(trn,val) in enumerate(self.cv_conf['iter']):
            trn_X,trn_y = self.train.loc[trn,self.features],self.train.loc[trn,self.label]
            val_X,val_y = self.train.loc[val,self.features],self.train.loc[val,self.label]

            estimator = CatBoostClassifier(**cat_params)
            print(cat_params)
            if os.path.exists(save_path+f'/model_cbt_f{n}.cbt'):
                estimator.load_model(save_path+f'/model_cbt_f{n}.cbt')
            else:
                estimator.fit(
                    trn_X,trn_y,
                    cat_features=cat_features,
                    # early_stopping_rounds=100,
                    # eval_set=[(val_X,val_y)],
                    eval_set=[(val_X,val_y)] if cat_params['task_type'] == 'GPU' else [(trn_X,trn_y),(val_X,val_y)],
                    use_best_model=True,
                    metric_period=200,
                    verbose=True,
                )
                estimator.save_model(save_path+f'/model_cbt_f{n}.cbt')

            oof_predictions[val] = estimator.predict_proba(val_X)[:,1]
            self.model.append(estimator)
            cv_score_list.append(roc_auc_score(y_true=val_y,y_score=estimator.predict_proba(val_X)[:,1]))
            tt_predicts += estimator.predict_proba(self.test[self.features])[:,1] / self.cv_conf['n']

            if self.run is not None:
                self.run['metrics/test_auc'].log(cv_score_list[-1])

        print(f"training CV oof mean : {np.round(np.mean(cv_score_list), 5)}")
        self.__check_diff_score(oof_predictions,val_idx=val)
        self.predictions = tt_predicts
        # self.features_imp = feature_imp
        if self.run is not None:
            self.run['feature_importance'] = self.features_importance()
        if cv_score:
            return oof_predictions,tt_predicts,cv_score_list
        else:
            if self.cv_conf['n']==1:
                # print(oof_predictions.shape,oof_predictions[val].shape,)
                return oof_predictions[val],tt_predicts,val_y.values

            return oof_predictions,tt_predicts

    def sklearn_test(self, sklean_model, cv_score=False, weight=None):
        # self.__deal_cat_features(cat_features,m_type='lgb')
        # imp = pd.DataFrame({
        #     'feature':self.features,
        #     'gain':0,
        #     'split':0,
        # })

        feature_imp = np.zeros(len(self.features))
        cv_score_list = []
        oof_predictions = np.zeros(len(self.train))
        tt_predicts = np.zeros(len(self.test))

        for n, (trn, val) in enumerate(self.cv_conf['iter']):
            trn_X, trn_y = self.train.loc[trn, self.features], self.train.loc[trn, self.label]
            val_X, val_y = self.train.loc[val, self.features], self.train.loc[val, self.label]


            estimator = sklean_model.fit(trn_X,trn_y)

            oof_predictions[val] = estimator.predict_proba(val_X)[:,1]
            self.model.append(estimator)

            cv_score_list.append(roc_auc_score(y_true=val_y, y_score=oof_predictions[val]))
            tt_predicts += estimator.predict_proba(self.test[self.features])[:,1] / self.cv_conf['n']
            
            if self.run is not None:
                self.run['metrics/test_auc'].log(cv_score_list[-1])


        print(f"training CV oof mean : {np.round(np.mean(cv_score_list), 5)}")

        self.__check_diff_score(oof_predictions)
        self.predictions = tt_predicts
        self.features_imp = feature_imp

        if self.run is not None:
            self.run['feature_importance'] = self.features_importance()
        if cv_score:
            return oof_predictions,tt_predicts,cv_score_list
        else:
            if self.cv_conf['n']==1:
                # print(oof_predictions.shape,oof_predictions[val].shape,)
                return oof_predictions[val],tt_predicts,val_y.values

            return oof_predictions,tt_predicts

    def find_threshold(self,oof,ylabel,weight_range,method='offline',
    predictions=None,online_thd=None):
        print(method)
        if method == 'offline':
            res = pd.DataFrame({
                'weight': weight_range,
                'oof_f1': 0,
                # 'oof_auc': 0,

            })
            # def search_weight(raw_prob, init_weight=[1.0]*n_classes, step=0.001):
            weight = weight_range.copy()*self.num_class
            step = 0.001
            f_best = accuracy_score(y_pred =np.argmax(oof,axis=1),y_true=ylabel)
            flag_score = 0
            round_num = 1
            # while(f_best < 1.77 and round_num < 5):
            while round_num < 5:
                print('round: ', round_num)
                round_num += 1
                flag_score = f_best
                for c in tqdm(range(self.num_class)):
                    for n_w in range(400, 1400, 10):
                        num = n_w * step
                        new_weight = weight.copy()
                        new_weight[c] = num

                        prob_df = oof.copy()
                        prob_df = prob_df * np.array(new_weight)

                        f = accuracy_score(y_pred=np.argmax(prob_df,axis=1),y_true=ylabel)
                        if f > f_best:
                            weight = new_weight.copy()
                            f_best = f
                            print(f_best)
                            # print(weight)
                            
            return weight




            for n in weight_range:
                oof_f1 = f1_score(y_true=ylabel,y_pred=[int(x>n) for x in oof ])
                # oof_auc = roc_auc_score(y_score=oof,y_true=ylabel)

                # print('the weighted oof auc : ',oof_auc)
                print(f'the weighted {np.round(n,4)} oof f1 {np.round(oof_f1,4)} ')
                # res.loc[res['weight'] == n,'oof_auc'] = oof_auc
                res.loc[res['weight'] == n,'oof_f1'] = oof_f1
            fig = plt.figure(figsize=(8, 6))
            ax2 = fig.add_subplot(111)
            res.index = res['weight']

            res['oof_f1'].plot(ax=ax2, style='y>-.', alpha=0.7, legend=True)
            plt.legend(loc=1)
            plt.show()
            best_threshold = res.loc[res['oof_f1']==res['oof_f1'].max(),'weight'].values
            print(f'the best threshold {best_threshold} and best f1 {res["oof_f1"].max()}')



        elif method == 'ythreshold':

            print(f'train dataset 01 distributions :{sum(ylabel)/len(ylabel)}')
            preds_index = int(len(predictions)*(1-(sum(ylabel)/len(ylabel))))
            sorted_pred = sorted(predictions)
            best_threshold = sorted_pred[preds_index]

            ##
            oof_index = sum(ylabel)
            sorted_oof = sorted(oof)
            best_oof = sorted_oof[oof_index]

            res = f1_score(y_true=ylabel,y_pred=[int(x>=best_oof) for x in oof],average='macro')
            print(f'weight is {best_oof} and test weight {best_threshold} oof was {res}')
        elif method == 'online':
            preds_index = int(len(predictions)*(online_thd/(1-online_thd)))
            sorted_pred = sorted(predictions)
            best_threshold = sorted_pred[preds_index]
            print(f'online wight is {best_threshold}')
            res = 0

        elif method == 'sortedPvalue':
            best_threshold,res = 0,0
            predictions = sorted(list(predictions))

            plt.figure(figsize=(20,10))
            sns.lineplot(predictions,)
            plt.show()
        else:
            raise NotImplementedError

        return best_threshold,res

    def features_importance(self):
        df = pd.DataFrame(self.features_imp, index=self.features,columns=['values'])
        df.sort_values('values',ascending=False,inplace=True,)

        return df

    def analysis_residual(self,oof_predictions):
        self.train['predic_res'] = oof_predictions
        res_train_t = self.train[self.train[self.label] == np.round(oof_predictions)].copy()
        res_train_f = self.train[self.train[self.label] != np.round(oof_predictions)].copy()
        try:
            import os
            os.mkdir('res_analysis')
        except:
            pass
        today = time.strftime("%Y-%m-%d", time.localtime())[5:]
        res_train_f.to_csv(f'./res_analysis/res_train_f_{today}.csv',index=False)
        res_train_t.to_csv(f'./res_analysis/res_train_t_{today}.csv',index=False)

    def submit_seeds_ensembel(self,lgb_params,seed_list):
        """
        the aim of fuction is to find good seed or ensembel seek to submit.

        """
        tt_predictions = np.zeros(len(self.test))
        seed_df = pd.DataFrame({
            'seed':seed_list,
            'oof_score':0,
            'cv_mean':0,
            'cv_std':0,
        })
        final_oof = np.zeros(len(self.train))

        for n,s in enumerate(seed_list):
            # seed_everything(s)
            self.init_CV(seed=s)
            oof_preds,tt_preds,cv_list = self.lgb_test(lgb_params=lgb_params,cv_score=True)
            tt_predictions += tt_preds/len(seed_list)
            final_oof += oof_preds/len(seed_df)

            seed_df.loc[n,'oof_auc'] = roc_auc_score(y_score=oof_preds,y_true=self.train[self.label])
            seed_df.loc[n,'oof_score'] = tpr_weight_funtion(y_predict=oof_preds,y_true=self.train[self.label])
            seed_df.loc[n,'cv_mean'] = np.mean(cv_list)
            seed_df.loc[n,'cv_std'] = np.std(cv_list)
            print(seed_df.loc[n,:],end='\n')

        self.predictions = tt_predictions
        global_auc = roc_auc_score(y_true=self.train[self.label],y_score=final_oof)
        global_tpr = tpr_weight_funtion(y_true=self.train[self.label],y_predict=final_oof)
        self.m_score.append([global_auc,global_tpr])
        seed_df['diff_abs'] = (seed_df['oof_auc'] - seed_df['cv_mean']).abs()
        seed_df.sort_values(['diff_abs','cv_std'],inplace=True)
        return seed_df

    def submit_pseudo_label_model(self,params,seed,model_type):
        # if len(self.predictions) == 0:
        print('training pseudo label .....')
        self.test['label'] = self.predictions.copy()
        test2p = self.test[(self.test['label']<=0.01)|(self.test['label']>=0.95)].copy()
        # relabel
        test2p.loc[test2p['label'] >= 0.5,'label'] = 1
        test2p.loc[test2p['label'] < 0.5,'label'] = 0

        train2p = pd.concat([self.train,test2p],axis=0)
        train2p.reset_index(drop=True,inplace=True)

        tmp_mt = make_test(train2p,self.test,self.features,[],m_score=self.m_score,label=self.label)
        tmp_mt.init_CV(seed)
        if model_type =='lgb':
            tmp_mt.lgb_test(params)
        if model_type == 'xgb':
            tmp_mt.xgb_test(params)

        self.predictions = tmp_mt.predictions

    def submit(self,ID,sub_file=True,threshold=None):
        today = time.strftime("%Y-%m-%d", time.localtime())[5:]
        # self.test[self.label] = [int(x) for x in self.predictions > 0.5]
        if threshold is None:
            self.test[self.label] = self.predictions
        else:
            self.test[self.label] = [int(x >= threshold) for x in self.predictions]
            print('sum of label',np.sum(self.test[self.label]))

        sub_train = self.test[[ID,self.label]].copy()
        if sub_file:
            sub_test = pd.read_csv('./data/submit.csv')
            sub = sub_test[[ID]].merge(sub_train, on=ID, how='left')
        else:
            sub = sub_train.copy()

        # plt.figure(figsize=(20, 10))
        # sns.distplot(sub['label'], bins=100)
        # plt.show()
        print('null in sub','\n',sub.isnull().sum())
        sub.fillna(0, inplace=True)
        score = str(np.round(self.m_score[-1][0],4))+"_"+str(np.round(self.m_score[-1][1],4))
        sub.to_csv(f'./result/sub_{today}_{score}.csv', index=False)
        return sub

    def psm_samples(self,lgb_params):

        ## p值
        tr,tt = self.train.copy(),self.test.copy()
        tr['test_label'] = 0
        tt['test_label'] = 1
        data = pd.concat([tr,tt],ignore_index=True,)
        # train,test = train_test_split(data,test_size=0.1,random_state=0)

        mt = make_test(data,data,self.features,[],m_score=[[0,0,0]],label='test_label')
        mt.init_CV(0)
        oof , test = mt.lgb_test(lgb_params)
        data['oof'] = oof
        ## 可视化
        # tmp_df = self.train[[self.label,'oof']]
        sns.boxplot(x='label',y='oof',data=data)
        plt.show()
        ## 找特征重要性
        imp_df = mt.features_importance()
        imp_df.reset_index(drop=False,inplace=True)
        imp_df.columns = ['features','values']
        sns.barplot(y='features',x='values',data=imp_df)
        plt.show()
        print(imp_df.head(10))
        return data,imp_df

    def submit_easy_ensemble(self,n_group=10,):
        pass

    def tune_model(self,params,model,type='binary',N_TRIALS=100,model_type='lgb'):

        def objective(trial):
            if model_type == 'lgb':
                obj_params = {
                    # 'random_state': params['seed'],
                    'metric': params['metric'],
                    'n_estimators': 5000,
                    'n_jobs': -1,
                    'seed':params['seed'],
                    'early_stopping_rounds': 100,
                    'learning_rate': trial.suggest_float('learning_rate', 1e-4, 0.1),
                    'max_depth': trial.suggest_int('max_depth', 6, 16),
                    'num_leaves': trial.suggest_int('num_leaves', 31, 2**8),
                    'reg_alpha': trial.suggest_float('reg_alpha', 1e-3, 10.0),
                    'reg_lambda': trial.suggest_float('reg_lambda', 1e-3, 10.0),
                    'colsample_bytree': trial.suggest_float('colsample_bytree', 0.2, 0.9),
                    'min_child_samples': trial.suggest_int('min_child_samples', 1, 300),
                    'subsample_freq': trial.suggest_int('subsample_freq', 1, 10),
                    'subsample': trial.suggest_float('subsample', 0.3, 0.9),
                    'max_bin': trial.suggest_int('max_bin', 128, 1024),
                    'min_data_per_group': trial.suggest_int('min_data_per_group', 50, 200),
                    'cat_smooth': trial.suggest_int('cat_smooth', 10, 100),
                    'cat_l2': trial.suggest_int('cat_l2', 1, 20),
                    'verbose': -1,
                      }

            elif model_type == 'xgb':
                obj_params = {
                    'objective': params['objective'],
                    # 'task_type': 'GPU',
                    'nthread': -1,
                    'seed': params['seed'],
                    'tree_method': 'hist',
                    'eval_metric': params['eval_metric'],
                    'lambda': trial.suggest_loguniform('lambda', 1e-3, 10.0),
                    'alpha': trial.suggest_loguniform('alpha', 1e-3, 10.0),
                    'colsample_bytree': trial.suggest_categorical('colsample_bytree', [0.3,0.4,0.5,0.6,0.7,0.8,0.9, 1.0]),
                    'subsample': trial.suggest_categorical('subsample', [0.4,0.5,0.6,0.7,0.8,1.0]),
                    'learning_rate': trial.suggest_categorical('learning_rate', [0.008,0.009,0.01,0.012,0.014,0.016,0.018, 0.02]),
                    # 'n_estimators': 10000,
                    'max_depth': trial.suggest_int('max_depth', 5,14),
                    # 'min_child_weight': trial.suggest_int('min_child_weight', 1, 300),
                }

            else:
                raise NotImplementedError
            oof, _ = model(obj_params, cv_score=False,)
            score = 0
            if type == 'binary':
                score = roc_auc_score(y_true=self.train[self.label],y_score=oof)
            if type == 'f1':
                score = f1_score(y_true=self.train[self.label],y_pred=[int(x>0.5) for x in oof],average='macro')

            return score

        study = optuna.create_study(study_name=f"optimization", direction='maximize')
        study.optimize(objective, n_trials=N_TRIALS)
        print('Number of finished trials:', len(study.trials))
        print('Best trial:', study.best_trial.params)
        ## save log
        print(study.trials_dataframe())
        import os
        if not os.path.exists('user_data'):
            os.mkdir('user_data')

        study.trials_dataframe().to_csv('./user_data/' + f"{type}trial_parameters.csv", index=False)
        with open('./user_data/' + f'{type}_study.pickle', 'wb') as f:
            pickle.dump(study, f)

        fig = plot_param_importances(study,)
        plotly.offline.plot(fig, filename='imp_fig.html')
        fig = plot_optimization_history(study)
        plotly.offline.plot(fig, filename='history_fig.html')

        return study.best_trial.params
